from nibabel import load as nload
from wonambi.attr import Channels, Freesurfer
from wonambi.attr.chan import (find_channel_groups,
                               create_sphere_around_elec,
                               )

from .paths import (chan_path,
                    fs_path,
                    template_mri_path,
                    exported_chan_path,
                    )

chan = Channels(chan_path)

def test_channels():
    assert chan.n_chan == 28

    xyz = chan.return_xyz()
    assert xyz.shape == (28, 3)

    labels = chan.return_label()
    assert len(labels) == 28


def test_channels_grid():
    grid_chan = chan(lambda x: x.label.startswith('grid'))
    assert grid_chan.n_chan == 16

    grid_chan.export(exported_chan_path)

    imported = Channels(exported_chan_path)
    assert imported.n_chan == grid_chan.n_chan


def test_channel_groups():
    groups = find_channel_groups(chan)
    assert len(groups) == 3


def test_channel_sphere():
    xyz = chan.return_xyz()[0,:]
    fs = Freesurfer(fs_path)
    mask = create_sphere_around_elec(xyz, template_mri_path,
                                     distance=8,
                                     freesurfer=fs)
    assert mask.sum() == 4

    template_mri = nload(str(template_mri_path))

    xyz_volume = xyz + fs.surface_ras_shift
    mask = create_sphere_around_elec(xyz, template_mri, distance=16)
    assert mask.sum() == 35



from os import SEEK_SET, SEEK_END, SEEK_CUR
from datetime import datetime, date
from struct import unpack

from numpy import array, dtype, empty, fromfile, iinfo, memmap, NaN, pad

N_ZONES = 15
MAX_SAMPLE = 128
MAX_CAN_VIEW = 128


class Micromed:
    """Basic class to read Micromed data. Supports .TRC and .VWR files

    Parameters
    ----------
    filename : path to file
        the name of the filename or directory
    """
    def __init__(self, filename):
        self.filename = filename
        self._header = {}

        with open(self.filename, 'rb') as f:
            self._header = _read_header(f)

            f.seek(0, SEEK_END)
            EOData = f.tell()

        self._bodata = self._header['BOData']
        self._n_chan = self._header['n_chan']
        self._n_bytes = self._header['n_bytes']
        self._s_freq = self._header['s_freq']
        self._n_smp = int((EOData - self._bodata) /
                          (self._n_chan * self._n_bytes))

        self._factors = array([ch['factor'] for ch in self._header['chans']])
        self._offset = array([ch['logical_ground'] for ch in self._header['chans']])

        self._triggers = self._header['trigger']
        self._videos = self._header['dvideo']

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header
        """

        subj_id = self._header['name'] + ' ' + self._header['surname']
        chan_name = [ch['chan_name'] for ch in self._header['chans']]

        return subj_id, self._header['start_time'], self._header['s_freq'], chan_name, self._n_smp, self._header

    def return_dat(self, chan, begsam, endsam):
        """Return the data as 2D numpy.ndarray.

        Parameters
        ----------
        chan : int or list
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, with dimension chan X samples
        """

        if type(chan) == int:  # if single value is provided it needs to be transformed to list to generate a 2d matrix
            chan = [chan, ]

        if (begsam >= self._n_smp) or (endsam < 0):
            dat = empty((len(chan), endsam - begsam))
            dat.fill(NaN)
            return dat

        if begsam < 0:
            begpad = -1 * begsam
            begsam = 0
        else:
            begpad = 0

        if endsam > self._n_smp:
            endpad = endsam - self._n_smp
            endsam = self._n_smp
        else:
            endpad = 0

        dshape = (self._n_chan, endsam - begsam)
        sig_dtype = 'u' + str(self._n_bytes)
        offset = self._bodata + begsam * self._n_bytes * self._n_chan
        dat = memmap(str(self.filename), dtype=sig_dtype, order='F', mode='r',
                     shape=dshape, offset=offset).astype('float')

        dat = pad(dat[chan, :], ((0, 0), (begpad, endpad)), mode='constant',
                  constant_values=NaN)

        return (dat - self._offset[chan, None]) * self._factors[chan, None]

    def return_markers(self):
        """Return all the markers (also called triggers or events).

        Returns
        -------
        list of dict
            where each dict contains 'name' as str, 'start' and 'end' as float
            in seconds from the start of the recordings, and 'chan' as list of
            str with the channels involved (if not of relevance, it's None).

        Raises
        ------
        FileNotFoundError
            when it cannot read the events for some reason (don't use other
            exceptions).
        """
        markers = []

        triggers = self._triggers
        DTYPE_MAX = iinfo(triggers.dtype['sample']).max
        triggers = triggers[triggers['sample'] != DTYPE_MAX]

        for trig in triggers:
            markers.append(
                {'name': str(trig['code']),
                 'start': trig['sample'] / self._s_freq,
                 'end': trig['sample'] / self._s_freq,
                 })

        return markers

    def return_videos(self, begtime, endtime):
        vid = self._videos

        # remove empty rows
        DTYPE_MAX = iinfo(vid.dtype['delay']).max
        vid = vid[vid['delay'] != DTYPE_MAX]

        if vid.shape[0] == 0:
            raise OSError('No videos for this dataset')

        if vid.shape[0] > 1:
            raise NotImplementedError('Currently it handles only one video.')

        # TODO: to test if delay is positive or negative
        video_beg = begtime + vid['delay'].item() / 1000
        video_end = endtime + vid['delay'].item() / 1000

        # full name without number
        video_name = 'VID_' + str(vid['file_ext'].item()) + '.AVI'

        mpgfiles = [
            str(self.filename.parent / video_name),
            ]

        return mpgfiles, video_beg, video_end


def _read_header(f):

    orig = {}

    orig['title'] = f.read(32).decode('utf-8').strip()
    orig['laboratory'] = f.read(32).strip(b'\x00').decode('utf-8').strip()

    # patient
    orig['surname'] = f.read(22).decode('utf-8').strip()
    orig['name'] = f.read(20).decode('utf-8').strip()
    month, day, year = unpack('bbb', f.read(3))
    orig['date_of_birth'] = date(year + 1900, month, day)
    f.seek(19, SEEK_CUR)

    # recording
    day, month, year, hour, minute, sec = unpack('bbbbbb', f.read(6))
    orig['start_time'] = datetime(year + 1900, month, day, hour, minute, sec)

    acquisition_unit_code = unpack('h', f.read(2))[0]
    orig['acquisition_unit'] = ACQUISITION_UNIT.get(acquisition_unit_code,
                                                    str(acquisition_unit_code))
    filetype_code = unpack('H', f.read(2))[0]
    orig['filetype'] = FILETYPE[filetype_code]

    orig['BOData'] = unpack('I', f.read(4))[0]
    orig['n_chan'] = unpack('H', f.read(2))[0]
    orig['multiplexer'] = unpack('H', f.read(2))[0]
    orig['s_freq'] = unpack('H', f.read(2))[0]
    orig['n_bytes'] = unpack('H', f.read(2))[0]
    orig['compression'] = unpack('H', f.read(2))[0]  # 0 non compression, 1 compression.
    orig['n_montages'] = unpack('H', f.read(2))[0]  # Montages : number of specific montages
    orig['dvideo_begin'] = unpack('I', f.read(4))[0]  # Starting sample of digital video
    orig['mpeg_delay'] = unpack('H', f.read(2))[0]  # Number of frames per hour of de-synchronization in MPEG acq

    f.seek(15, SEEK_CUR)
    header_type_code = unpack('b', f.read(1))[0]
    orig['header_type'] = HEADER_TYPE[header_type_code]

    zones = {}
    for _ in range(N_ZONES):
        zname, pos, length = unpack('8sII', f.read(16))
        zname = zname.decode('utf-8').strip()
        zones[zname] = pos, length

    pos, length = zones['ORDER']
    f.seek(pos, SEEK_SET)
    order = fromfile(f, dtype='u2', count=orig['n_chan'])

    chans = _read_labcod(f, zones['LABCOD'], order)

    pos, length = zones['NOTE']
    f.seek(pos, SEEK_SET)
    DTYPE = dtype([('sample', 'u4'), ('text', 'S40')])
    notes = fromfile(f, dtype=DTYPE, count=int(length / DTYPE.itemsize))

    pos, length = zones['FLAGS']
    f.seek(pos, SEEK_SET)
    DTYPE = dtype([('begin', 'u4'), ('end', 'u4')])
    flags = fromfile(f, dtype=DTYPE, count=int(length / DTYPE.itemsize))

    pos, length = zones['TRONCA']
    f.seek(pos, SEEK_SET)
    DTYPE = dtype([('time_in_samples', 'u4'), ('sample', 'u4')])
    segments = fromfile(f, dtype=DTYPE, count=int(length / DTYPE.itemsize))

    # impedance
    DTYPE = dtype([('positive', 'u1'), ('negative', 'u1')])
    pos, length = zones['IMPED_B']
    f.seek(pos, SEEK_SET)
    impedance_begin = fromfile(f, dtype=DTYPE, count=int(length / DTYPE.itemsize))

    pos, length = zones['IMPED_E']
    f.seek(pos, SEEK_SET)
    impedance_end = fromfile(f, dtype=DTYPE, count=int(length / DTYPE.itemsize))

    montage = _read_montage(f, zones['MONTAGE'])

    # if average has been computed
    pos, length = zones['COMPRESS']
    f.seek(pos, SEEK_SET)
    avg = {}
    avg['trace'], avg['file'], avg['prestim'], avg['poststim'], avg['type'] = unpack('IIIII', f.read(5 * 4))
    avg['free'] = f.read(108).strip(b'\x01\x00')

    history = _read_history(f, zones['HISTORY'])

    pos, length = zones['DVIDEO']
    f.seek(pos, 0)
    DTYPE = dtype([('delay', 'u4'), ('duration', 'u4'), ('file_ext', 'u4'), ('empty', 'u4')])
    dvideo = fromfile(f, dtype=DTYPE, count=int(length / DTYPE.itemsize))

    # events
    DTYPE = dtype([('code', 'u4'), ('begin', 'u4'), ('end', 'u4')])
    pos, length = zones['EVENT A']
    f.seek(pos, 0)
    event_a = fromfile(f, dtype=DTYPE, count=int(length / DTYPE.itemsize))

    pos, length = zones['EVENT B']
    f.seek(pos, 0)
    event_b = fromfile(f, dtype=DTYPE, count=int(length / DTYPE.itemsize))

    pos, length = zones['TRIGGER']
    DTYPE = dtype([('sample', 'u4'), ('code', 'u2')])
    f.seek(pos, 0)
    trigger = fromfile(f, dtype=DTYPE, count=int(length / DTYPE.itemsize))

    orig.update({
        'order': order,
        'chans': chans,
        'notes': notes,
        'flags': flags,
        'segments': segments,
        'impedance_begin': impedance_begin,
        'impedance_end': impedance_end,
        'montage': montage,
        'history': history,
        'dvideo': dvideo,
        'event_a': event_a,
        'event_b': event_b,
        'trigger': trigger,
        })

    return orig


ACQUISITION_UNIT = {
    0: 'BQ124 - 24 channels headbox, Internal Interface',
    2: 'MS40 - Holter recorder',
    6: 'BQ132S - 32 channels headbox, Internal Interface',
    7: 'BQ124 - 24 channels headbox, BQ CARD Interface',
    8: 'SAM32 - 32 channels headbox, BQ CARD Interface',
    9: 'SAM25 - 25 channels headbox, BQ CARD Interface',
    10: 'BQ132S R - 32 channels reverse headbox, Internal Interface',
    11: 'SAM32 R - 32 channels reverse headbox, BQ CARD Interface',
    12: 'SAM25 R - 25 channels reverse headbox, BQ CARD Interface',
    13: 'SAM32 - 32 channels headbox, Internal Interface',
    14: 'SAM25 - 25 channels headbox, Internal Interface',
    15: 'SAM32 R - 32 channels reverse headbox, Internal Interface',
    16: 'SAM25 R - 25 channels reverse headbox, Internal Interface',
    17: 'SD - 32 channels headbox with jackbox, SD CARD Interface -- PCI Internal Interface',
    18: 'SD128 - 128 channels headbox, SD CARD Interface -- PCI Internal Interface',
    19: 'SD96 - 96 channels headbox, SD CARD Interface -- PCI Internal Interface',
    20: 'SD64 - 64 channels headbox, SD CARD Interface -- PCI Internal Interface',
    21: 'SD128c - 128 channels headbox with jackbox, SD CARD Interface -- PCI Internal Interface',
    22: 'SD64c - 64 channels headbox with jackbox, SD CARD Interface -- PCI Internal Interface',
    23: 'BQ132S - 32 channels headbox, PCI Internal Interface',
    24: 'BQ132S R - 32 channels reverse headbox, PCI Internal Interface',
    }

FILETYPE = {
    40: 'C128 C.R., 128 EEG (headbox SD128 only)',
    42: 'C84P C.R., 84 EEG, 44 poly (headbox SD128 only)',
    44: 'C84 C.R., 84 EEG, 4 reference signals (named MKR,MKRB,MKRC,MKRD) (headbox SD128 only)',
    46: 'C96 C.R., 96 EEG (headbox SD128 -- SD96 -- BQ123S(r))',
    48: 'C63P C.R., 63 EEG, 33 poly',
    50: 'C63 C.R., 63 EEG, 3 reference signals (named MKR,MKRB,MKRC)',
    52: 'C64 C.R., 64 EEG',
    54: 'C42P C.R., 42 EEG, 22 poly',
    56: 'C42 C.R., 42 EEG, 2 reference signals (named MKR,MKRB)',
    58: 'C32 C.R., 32 EEG',
    60: 'C21P C.R., 21 EEG, 11 poly',
    62: 'C21 C.R., 21 EEG, 1 reference signal (named MKR)',
    64: 'C19P C.R., 19 EEG, variable poly',
    66: 'C19 C.R., 19 EEG, 1 reference signal (named MKR)',
    68: 'C12 C.R., 12 EEG',
    70: 'C8P C.R., 8 EEG, variable poly',
    72: 'C8 C.R., 8 EEG',
    74: 'CFRE C.R., variable EEG, variable poly',
    76: 'C25P C.R., 25 EEG (21 standard, 4 poly transformed to EEG channels), 7 poly -- headbox BQ132S(r) only',
    78: 'C27P C.R., 27 EEG (21 standard, 6 poly transformed to EEG channels), 5 poly -- headbox BQ132S(r) only',
    80: 'C24P C.R., 24 EEG (21 standard, 3 poly transformed to EEG channels), 8 poly -- headbox SAM32(r) only',
    82: 'C25P C.R., 25 EEG (21 standard, 4 poly transformed to EEG channels), 7 poly -- headbox SD with headbox JB 21P',
    84: 'C27P C.R., 27 EEG (21 standard, 6 poly transformed to EEG channels), 5 poly -- headbox SD with headbox JB 21P',
    86: 'C31P C.R., 27 EEG (21 standard, 10 poly transformed to EEG channels), 1 poly -- headbox SD with headbox JB 21P6',
    100: 'C26P C.R., 26 EEG, 6 poly (headbox SD, SD64c, SD128c with headbox JB Mini)',
    101: 'C16P C.R., 16 EEG, 16 poly (headbox SD with headbox JB M12)',
    102: 'C12P C.R., 12 EEG, 20 poly (headbox SD with headbox JB M12)',
    103: '32P 32 poly (headbox SD, SD64c, SD128c with headbox JB Bip)',
    120: 'C48P C.R., 48 EEG, 16 poly (headbox SD64)',
    121: 'C56P C.R., 56 EEG, 8 poly (headbox SD64)',
    122: 'C24P C.R., 24 EEG, 8 poly (headbox SD64)',
    140: 'C52P C.R., 52 EEG, 12 poly (headbox SD64c, SD128c with 2 headboxes JB Mini)',
    141: '64P 64 poly (headbox SD64c, SD128c with 2 headboxes JB Bip)',
    160: 'C88P C.R., 88 EEG, 8 poly (headbox SD96)',
    161: 'C80P C.R., 80 EEG, 16 poly (headbox SD96)',
    162: 'C72P C.R., 72 EEG, 24 poly (headbox SD96)',
    180: 'C120P C.R., 120 EEG, 8 poly (headbox SD128)',
    181: 'C112P C.R., 112 EEG, 16 poly (headbox SD128)',
    182: 'C104P C.R., 104 EEG, 24 poly (headbox SD128)',
    183: 'C96P C.R., 96 EEG, 32 poly (headbox SD128)',
    200: 'C122P C.R., 122 EEG, 6 poly (headbox SD128c with 4 headboxes JB Mini)',
    201: 'C116P C.R., 116 EEG, 12 poly (headbox SD128c with 4 headboxes JB Mini)',
    202: 'C110P C.R., 110 EEG, 18 poly (headbox SD128c with 4 headboxes JB Mini)',
    203: 'C104P C.R., 104 EEG, 24 poly (headbox SD128c with 4 headboxes JB Mini)',
    204: '128P 128 poly (headbox SD128c with 4 headboxes JB Bip)',
    205: '96P 96 poly (headbox SD128c with 3 headboxes JB Bip)',
    }

HEADER_TYPE = {
    0: 'Micromed "System 1" Header type',
    1: 'Micromed "System 1" Header type',
    2: 'Micromed "System 2" Header type',
    3: 'Micromed "System98" Header type',
    4: 'Micromed "System98" Header type',
    }

UNITS = {
    -1: 'nV',
    0: 'μV',
    1: 'mV',
    2: 'V',
    100: '%',
    101: 'bpm',
    102: 'dimentionless',
    }


def _read_labcod(f, zone, order):
    pos, length = zone
    CHAN_LENGTH = 128

    chans = []

    for i_ch in order:
        chan = {}

        f.seek(pos + i_ch * CHAN_LENGTH, 0)

        chan['status'] = f.read(1)  # Status of electrode for acquisition : 0 : not acquired, 1 : acquired
        chan['channelType'] = f.read(1)  # TODO: type of reference

        chan['chan_name'] = f.read(6).strip(b'\x01\x00').decode()
        chan['ground'] = f.read(6).strip(b'\x01\x00').decode()
        l_min, l_max, chan['logical_ground'], ph_min, ph_max = unpack('iiiii', f.read(20))
        chan['factor'] = float(ph_max - ph_min) / float(l_max - l_min + 1)

        k = unpack('h', f.read(2))[0]
        chan['units'] = UNITS.get(k, UNITS[0])

        chan['HiPass_Limit'], chan['HiPass_Type'] = unpack('HH', f.read(4))
        chan['LowPass_Limit'], chan['LowPass_Type'] = unpack('HH', f.read(4))

        chan['rate_coefficient'], chan['position'] = unpack('HH', f.read(4))
        chan['Latitude'], chan['Longitude'] = unpack('ff', f.read(8))
        chan['presentInMap'] = unpack('B', f.read(1))[0]
        chan['isInAvg'] = unpack('B', f.read(1))[0]
        chan['Description'] = f.read(32).strip(b'\x01\x00').decode()
        chan['xyz'] = unpack('fff', f.read(12))
        chan['Coordinate_Type'] = unpack('H', f.read(2))[0]
        chan['free'] = f.read(24).strip(b'\x01\x00')

        chans.append(chan)

    return chans


def _read_montage(f, zone):
    pos, length = zone
    f.seek(pos, SEEK_SET)

    montages = []

    while f.tell() < (pos + length):
        montage = {
            'lines': unpack('H', f.read(2))[0],
            'sectors': unpack('H', f.read(2))[0],
            'base_time': unpack('H', f.read(2))[0],
            'notch': unpack('H', f.read(2))[0],
            'colour': unpack(MAX_CAN_VIEW * 'B', f.read(MAX_CAN_VIEW)),
            'selection': unpack(MAX_CAN_VIEW * 'B', f.read(MAX_CAN_VIEW)),
            'description': f.read(64).strip(b'\x01\x00').decode(),
            'inputsNonInv': unpack(MAX_CAN_VIEW * 'H', f.read(MAX_CAN_VIEW * 2)),  # NonInv : non inverting input
            'inputsInv': unpack(MAX_CAN_VIEW * 'H', f.read(MAX_CAN_VIEW * 2)),  # Inv : inverting input

            'HiPass_Filter': unpack(MAX_CAN_VIEW * 'I', f.read(MAX_CAN_VIEW * 4)),
            'LowPass_Filter': unpack(MAX_CAN_VIEW * 'I', f.read(MAX_CAN_VIEW * 4)),
            'reference': unpack(MAX_CAN_VIEW * 'I', f.read(MAX_CAN_VIEW * 4)),
            'free': f.read(1720).strip(b'\x01\x00'),
            }
        montages.append(montage)

    return montages


def _read_history(f, zone):
    """This matches the Matlab reader from Matlab Exchange but doesn't seem
    correct.
    """
    pos, length = zone
    f.seek(pos, SEEK_SET)

    histories = []
    while f.tell() < (pos + length):
        history = {
            'nSample': unpack(MAX_SAMPLE * 'I', f.read(MAX_SAMPLE * 4)),
            'lines': unpack('H', f.read(2)),
            'sectors': unpack('H', f.read(2)),
            'base_time': unpack('H', f.read(2)),
            'notch': unpack('H', f.read(2)),
            'colour': unpack(MAX_CAN_VIEW * 'B', f.read(MAX_CAN_VIEW)),
            'selection': unpack(MAX_CAN_VIEW * 'B', f.read(MAX_CAN_VIEW)),
            'description': f.read(64).strip(b'\x01\x00'),
            'inputsNonInv': unpack(MAX_CAN_VIEW * 'H', f.read(MAX_CAN_VIEW * 2)),  # NonInv : non inverting input
            'inputsInv': unpack(MAX_CAN_VIEW * 'H', f.read(MAX_CAN_VIEW * 2)),  # Inv : inverting input

            'HiPass_Filter': unpack(MAX_CAN_VIEW * 'I', f.read(MAX_CAN_VIEW * 4)),
            'LowPass_Filter': unpack(MAX_CAN_VIEW * 'I', f.read(MAX_CAN_VIEW * 4)),
            'reference': unpack(MAX_CAN_VIEW * 'I', f.read(MAX_CAN_VIEW * 4)),
            'free': f.read(1720).strip(b'\x01\x00'),
            }
        histories.append(history)

    return histories



"""Various functions used for the GUI.
"""
from ast import literal_eval
from logging import getLogger
from math import ceil, floor
from numpy import arange
from os.path import dirname, join, realpath

from PyQt5.QtCore import QRectF, QSettings, Qt
from PyQt5.QtGui import (QBrush,
                         QColor,
                         QPainterPath,
                         QPainter,
                         )
from PyQt5.QtSvg import QSvgGenerator
from PyQt5.QtWidgets import (QCheckBox,
                             QComboBox,
                             QCommonStyle,
                             QFileDialog,
                             QGraphicsItem,
                             QGraphicsRectItem,
                             QGraphicsSimpleTextItem,
                             QLineEdit,
                             QMessageBox,
                             QPushButton,
                             QRadioButton,
                             QSpinBox,
                             )


lg = getLogger(__name__)

LINE_WIDTH = 0  # COSMETIC LINE
LINE_COLOR = 'black'
# TODO: this in ConfigNotes
STAGE_NAME = ['NREM1', 'NREM2', 'NREM3', 'REM', 'Wake', 'Movement',
              'Undefined', 'Unknown', 'Artefact']

MAX_LENGTH = 20

stdicon = QCommonStyle.standardIcon


icon_path = join(dirname(realpath(__file__)), 'icons')
oxy_path = join(icon_path, 'oxygen')

ICON = {'application': join(icon_path, 'wonambi.jpg'),
        'open_rec': join(oxy_path, 'document-open.png'),
        'page_prev': join(oxy_path, 'go-previous-view.png'),
        'page_next': join(oxy_path, 'go-next-view.png'),
        'step_prev': join(oxy_path, 'go-previous.png'),
        'step_next': join(oxy_path, 'go-next.png'),
        'chronometer': join(oxy_path, 'chronometer.png'),
        'up': join(oxy_path, 'go-up.png'),
        'down': join(oxy_path, 'go-down.png'),
        'zoomin': join(oxy_path, 'zoom-in.png'),
        'zoomout': join(oxy_path, 'zoom-out.png'),
        'zoomnext': join(oxy_path, 'zoom-next.png'),
        'zoomprev': join(oxy_path, 'zoom-previous.png'),
        'ydist_more': join(oxy_path, 'format-line-spacing-triple.png'),
        'ydist_less': join(oxy_path, 'format-line-spacing-normal.png'),
        'selchan': join(oxy_path, 'mail-mark-task.png'),
        'widget': join(oxy_path, 'window-duplicate.png'),
        'settings': join(oxy_path, 'configure.png'),
        'quit': join(oxy_path, 'window-close.png'),
        'bookmark': join(oxy_path, 'bookmarks-organize.png'),
        'event': join(oxy_path, 'edit-table-cell-merge.png'),
        'new_eventtype': join(oxy_path, 'edit-table-insert-column-right.png'),
        'del_eventtype': join(oxy_path, 'edit-table-delete-column.png'),
        'help-about': join(oxy_path, 'help-about.png')
        }

settings = QSettings("wonambi", "wonambi")


class Path(QPainterPath):
    """Paint a line in the simplest possible way.

    Parameters
    ----------
    x : ndarray or list
        x-coordinates
    y : ndarray or list
        y-coordinates
    """
    def __init__(self, x, y):
        super().__init__()

        self.moveTo(x[0], y[0])
        for i_x, i_y in zip(x, y):
            self.lineTo(i_x, i_y)


class RectMarker(QGraphicsRectItem):
    """Class to draw a rectangular, coloured item.

    Parameters
    ----------
    x : float
        x position in scene
    y : gloat
        y position in scene
    width : float
        length in seconds
    height : float
        height in scene units
    color : str or QColor, optional
        color of the rectangle
    """
    def __init__(self, x, y, width, height, zvalue, color='blue'):
        super().__init__()

        self.color = color
        self.setZValue(zvalue)
        buffer = 1
        self.marker = QRectF(x, y, width, height)
        self.b_rect = QRectF(x - buffer / 2, y + buffer / 2, width + buffer,
                             height + buffer)
        self.params = x, y, width, height, zvalue, color

    def boundingRect(self):
        return self.b_rect

    def paint(self, painter, option, widget):
        color = QColor(self.color)
        painter.setBrush(QBrush(color))
        painter.setPen(Qt.NoPen)
        painter.drawRect(self.marker)
        super().paint(painter, option, widget)

    def contains(self, pos):
        return self.marker.contains(pos)


class TextItem_with_BG(QGraphicsSimpleTextItem):
    """Class to draw text with dark background (easier to read).

    Parameters
    ----------
    bg_color : str or QColor, optional
        color to use as background
    """
    def __init__(self, bg_color='black'):
        super().__init__()

        self.bg_color = bg_color
        self.setFlag(QGraphicsItem.ItemIgnoresTransformations)
        self.setBrush(QBrush(Qt.white))

    def paint(self, painter, option, widget):
        bg_color = QColor(self.bg_color)
        painter.setBrush(QBrush(bg_color))
        painter.drawRect(self.boundingRect())
        super().paint(painter, option, widget)

class FormBool(QCheckBox):
    """Subclass QCheckBox to have a more consistent API across widgets.

    Parameters
    ----------
    checkbox_label : str
        label next to checkbox

    """
    def __init__(self, checkbox_label):
        super().__init__(checkbox_label)

    def get_value(self, default=False):
        """Get the value of the QCheckBox, as boolean.

        Parameters
        ----------
        default : bool
            not used

        Returns
        -------
        bool
            state of the checkbox

        """
        return self.checkState() == Qt.Checked

    def set_value(self, value):
        """Set value of the checkbox.

        Parameters
        ----------
        value : bool
            value for the checkbox

        """
        if value:
            self.setCheckState(Qt.Checked)
        else:
            self.setCheckState(Qt.Unchecked)

    def connect(self, funct):
        """Call funct when user ticks the box.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.stateChanged.connect(funct)


class FormRadio(QRadioButton):
    """Subclass QRadioButton to have a more consistent API across widgets.

    Parameters
    ----------
    checkbox_label : str
        label next to checkbox

    """
    def __init__(self, checkbox_label):
        super().__init__(checkbox_label)

    def get_value(self, default=False):
        """Get the value of the QCheckBox, as boolean.

        Parameters
        ----------
        default : bool
            not used

        Returns
        -------
        bool
            state of the checkbox

        """
        return self.isChecked == Qt.Checked

    def set_value(self, value):
        """Set value of the checkbox.

        Parameters
        ----------
        value : bool
            value for the checkbox

        """
        if value:
            self.setChecked(Qt.Checked)
        else:
            self.setChecked(Qt.Unchecked)

    def connect(self, funct):
        """Call funct when user ticks the box.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.toggled.connect(funct)


class FormFloat(QLineEdit):
    """Subclass QLineEdit for float to have a more consistent API across
    widgets.

    """
    def __init__(self, default=None):
        super().__init__('')

        if default is not None:
            self.set_value(default)

    def get_value(self, default=0):
        """Get float from widget.

        Parameters
        ----------
        default : float
            default value for the parameter in case it fails

        Returns
        -------
        float
            the value in text or default

        """
        text = self.text()
        try:
            text = float(text)
        except ValueError:
            lg.debug('Cannot convert "' + str(text) + '" to float.' +
                     'Using default ' + str(default))
            text = default
            self.set_value(text)
        return text

    def set_value(self, value):
        """Set value of the float.

        Parameters
        ----------
        value : float
            value for the line edit

        """
        self.setText(str(value))

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.textEdited.connect(funct)


class FormInt(QLineEdit):
    """Subclass QLineEdit for int to have a more consistent API across widgets.

    """
    def __init__(self, default=None):
        super().__init__('')

        if default is not None:
            self.set_value(default)

    def get_value(self, default=0):
        """Get int from widget.

        Parameters
        ----------
        default : int
            default value for the parameter in case it fails

        Returns
        -------
        int
            the value in text or default

        """
        text = self.text()
        try:
            text = int(float(text))  # to convert values like 30.0
        except ValueError:
            lg.debug('Cannot convert "' + str(text) + '" to int. ' +
                     'Using default ' + str(default))
            text = default
            self.set_value(text)
        return text

    def set_value(self, value):
        """Set value of the int.

        Parameters
        ----------
        value : int
            value for the line edit

        """
        self.setText(str(value))

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.textEdited.connect(funct)


class FormList(QLineEdit):
    """Subclass QLineEdit for lists to have a more consistent API across
    widgets.

    """
    def __init__(self):
        super().__init__('')

    def get_value(self, default=None):
        """Get int from widget.

        Parameters
        ----------
        default : list
            list with widgets

        Returns
        -------
        list
            list that might contain int or str or float etc

        """
        if default is None:
            default = []

        try:
            text = literal_eval(self.text())
            if not isinstance(text, list):
                pass
                # raise ValueError

        except ValueError:
            lg.debug('Cannot convert "' + str(text) + '" to list. ' +
                     'Using default ' + str(default))
            text = default
            self.set_value(text)

        return text

    def set_value(self, value):
        """Set value of the list.

        Parameters
        ----------
        value : list
            value for the line edit

        """
        self.setText(str(value))

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.textEdited.connect(funct)


class FormStr(QLineEdit):
    """Subclass QLineEdit for str to have a more consistent API across widgets.

    """
    def __init__(self):
        super().__init__('')

    def get_value(self, default=''):
        """Get int from widget.

        Parameters
        ----------
        default : str
            not used

        Returns
        -------
        str
            the value in text

        """
        return self.text()

    def set_value(self, value):
        """Set value of the string.

        Parameters
        ----------
        value : str
            value for the line edit

        """
        self.setText(value)

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.textEdited.connect(funct)


class FormDir(QPushButton):
    """Subclass QPushButton for str to have a more consistent API across widgets.

    Notes
    -----
    It calls to open the directory three times, but I don't understand why

    """
    def __init__(self):
        super().__init__('')

    def get_value(self, default=''):
        """Get int from widget.

        Parameters
        ----------
        default : str
            not used

        Returns
        -------
        str
            the value in text

        """
        return self.text()

    def set_value(self, value):
        """Set value of the string.

        Parameters
        ----------
        value : str
            value for the line edit

        """
        self.setText(value)

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        Notes
        -----
        There is something wrong here. When you run this function, it calls
        for opening a directory three or four times. This is obviously wrong
        but I don't understand why this happens three times. Traceback did not
        help.

        """
        def get_directory():
            rec = QFileDialog.getExistingDirectory(self,
                                                   'Path to Recording'
                                                   ' Directory')
            if rec == '':
                return

            self.setText(rec)
            funct()

        self.clicked.connect(get_directory)

class FormMenu(QComboBox):
    """Subclass QComboBox for dropdown menus to have a more consistent API
    across widgets.

    Parameters
    ----------
    input_list: list of str
        items to include in the dropdown menu / combobox
    """
    def __init__(self, input_list):
        super().__init__()

        if input_list is not None:
            for i in input_list:
                self.addItem(i)

    def get_value(self, default=None):
        """Get selection from widget.

        Parameters
        ----------
        default : str
            str for use by widget

        Returns
        -------
        str
            selected item from the combobox

        """
        if default is None:
            default = ''

        try:
            text = self.currentText()

        except ValueError:
            lg.debug('Cannot convert "' + str(text) + '" to list. ' +
                     'Using default ' + str(default))
            text = default
            self.set_value(text)

        return text

    def set_value(self, value):
        """Set value of the list.

        Parameters
        ----------
        value : str
            value for the combobox

        """
        self.setCurrentText(str(value))

    def connect(self, funct):
        """Call funct when the selection was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.currentIndexChanged.connect(funct)

class FormSpin(QSpinBox):
    """Subclass QSpinBox for int to have a more consistent API across widgets.

    """
    def __init__(self, default=None, min_val=None, max_val=None, step=None):
        super().__init__()

        if default is not None:
            self.set_value(default)

        if min_val is not None:
            self.setMinimum(min_val)

        if max_val is not None:
            self.setMaximum(max_val)

        if step is not None:
            self.setSingleStep(step)

    def get_value(self, default=0):
        """Get int from widget.

        Parameters
        ----------
        default : int
            default value for the parameter in case it fails

        Returns
        -------
        int
            the value in text or default

        """
        text = self.value()
        try:
            text = int(float(text))
        except ValueError:
            lg.debug('Cannot convert "' + str(text) + '" to int. ' +
                     'Using default ' + str(default))
            text = default
            self.set_value(text)
        return text

    def set_value(self, value):
        """Set value of the int.

        Parameters
        ----------
        value : int
            value for the line edit

        """
        self.setValue(int(value))

    def connect(self, funct):
        """Call funct when the text was changed.

        Parameters
        ----------
        funct : function
            function that broadcasts a change.

        """
        self.valueChanged.connect(funct)

def keep_recent_datasets(max_dataset_history, info=None):
    """Keep track of the most recent recordings.

    Parameters
    ----------
    max_dataset_history : int
        maximum number of datasets to remember
    info : str, optional TODO
        path to file

    Returns
    -------
    list of str
        paths to most recent datasets (only if you don't specify
        new_dataset)
    """
    history = settings.value('recent_recordings', [])
    if isinstance(history, str):
        history = [history]

    if info is not None and info.filename is not None:
        new_dataset = info.filename

        if new_dataset in history:
            lg.debug(new_dataset + ' already present, will be replaced')
            history.remove(new_dataset)
        if len(history) > max_dataset_history:
            lg.debug('Removing last dataset ' + history[-1])
            history.pop()

        lg.debug('Adding ' + new_dataset + ' to list of recent datasets')
        history.insert(0, new_dataset)
        settings.setValue('recent_recordings', history)
        return None

    else:
        return history


def choose_file_or_dir():
    """Create a simple message box to see if the user wants to open dir or file

    Returns
    -------
    str
        'dir' or 'file' or 'abort'

    """
    question = QMessageBox(QMessageBox.Information, 'Open Dataset',
                           'Do you want to open a file or a directory?')
    dir_button = question.addButton('Directory', QMessageBox.YesRole)
    file_button = question.addButton('File', QMessageBox.NoRole)
    question.addButton(QMessageBox.Cancel)
    question.exec_()
    response = question.clickedButton()

    if response == dir_button:
        return 'dir'
    elif response == file_button:
        return 'file'
    else:
        return 'abort'


def short_strings(s, max_length=MAX_LENGTH):
    if len(s) > max_length:
        max_length -= 3  # dots
        start = ceil(max_length / 2)
        end = -floor(max_length / 2)
        s = s[:start] + '...' + s[end:]
    return s


def convert_name_to_color(s):
    """Convert any string to an RGB color.

    Parameters
    ----------
    s : str
        string to convert
    selection : bool, optional
        if an event is being selected, it's lighter

    Returns
    -------
    instance of QColor
        one of the possible color

    Notes
    -----
    It takes any string and converts it to RGB color. The same string always
    returns the same color. The numbers are a bit arbitrary but not completely.
    h is the baseline color (keep it high to have brighter colors). Make sure
    that the max module + h is less than 256 (RGB limit).

    The number you multiply ord for is necessary to differentiate the letters
    (otherwise 'r' and 's' are too close to each other).
    """
    h = 100
    v = [5 * ord(x) for x in s]
    sum_mod = lambda x: sum(x) % 100
    color = QColor(sum_mod(v[::3]) + h, sum_mod(v[1::3]) + h,
                   sum_mod(v[2::3]) + h)
    return color


def freq_from_str(freq_str):
    """Obtain frequency ranges from input string, either as list or dynamic
    notation.

    Parameters
    ----------
    freq_str : str
        String with frequency ranges, either as a list:
        e.g. [[1-3], [3-5], [5-8]];
        or with a dynamic definition: (start, stop, width, step).

    Returns
    -------
    list of tuple of float or None
        Every tuple of float represents a frequency band. If input is invalid,
        returns None.
    """
    freq = []
    as_list = freq_str[1:-1].replace(' ', '').split(',')

    try:
        if freq_str[0] == '[' and freq_str[-1] == ']':
            for i in as_list:
                one_band = i[1:-1].split('-')
                one_band = float(one_band[0]), float(one_band[1])
                freq.append(one_band)
    
        elif freq_str[0] == '(' and freq_str[-1] == ')':
    
            if len(as_list) == 4:
                start = float(as_list[0])
                stop = float(as_list[1])
                halfwidth = float(as_list[2]) / 2
                step = float(as_list[3])
                centres = arange(start, stop, step)
                for i in centres:
                    freq.append((i - halfwidth, i + halfwidth))
            else:
                return None
    
        else:
            return None           
    except:
        return None

    return freq


def export_graphics(MAIN, checked=False, test=None):

    from .modal_widgets import SVGDialog  # avoid circolar import

    if MAIN.info.filename is not None:
        filename = join(dirname(MAIN.info.filename), '*.svg')
    else:
        filename = None

    svg_dialog = SVGDialog(filename)

    if test is None:
        if not svg_dialog.exec():
            return
    else:
        svg_dialog.idx_file.setText(test)

    if svg_dialog.idx_list.currentText() == 'Traces':
        widget = MAIN.traces
    elif svg_dialog.idx_list.currentText() == 'Overview':
        widget = MAIN.overview

    svg_file = svg_dialog.idx_file.text()
    if not svg_file.endswith('.svg'):
        svg_file += '.svg'
    export_graphics_to_svg(widget, svg_file)


def export_graphics_to_svg(widget, filename):
    """Export graphics to svg

    Parameters
    ----------
    widget : instance of QGraphicsView
        traces or overview
    filename : str
        path to save svg
    """
    generator = QSvgGenerator()
    generator.setFileName(filename)
    generator.setSize(widget.size())
    generator.setViewBox(widget.rect())

    painter = QPainter()
    painter.begin(generator)
    widget.render(painter)
    painter.end()



"""Functions used when creating a new window.

"""
from logging import getLogger
from functools import partial

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (QAction,
                             QDockWidget,
                             QMessageBox,
                             )

from .settings import Settings # has to be first
from .utils import ICON   # has to be second
from .labels import Labels
from .channels import Channels
from .info import Info, ExportDatasetDialog
from .overview import Overview
from .notes import Notes, MergeDialog, ExportEventsDialog
from .detect_dialogs import SpindleDialog, SWDialog
from .analysis import AnalysisDialog
from .spectrum import Spectrum
from .traces import Traces
from .video import Video


lg = getLogger(__name__)


def create_widgets(MAIN):
    """Create all the widgets and dockwidgets. It also creates actions to
    toggle views of dockwidgets in dockwidgets.
    """

    """ ------ CREATE WIDGETS ------ """
    MAIN.labels = Labels(MAIN)
    MAIN.channels = Channels(MAIN)
    MAIN.notes = Notes(MAIN)
    MAIN.merge_dialog = MergeDialog(MAIN)
    MAIN.export_events_dialog = ExportEventsDialog(MAIN)
    MAIN.export_dataset_dialog = ExportDatasetDialog(MAIN)
    MAIN.spindle_dialog = SpindleDialog(MAIN)
    MAIN.slow_wave_dialog = SWDialog(MAIN)
    MAIN.analysis_dialog = AnalysisDialog(MAIN)
    #MAIN.plot_dialog = PlotDialog(MAIN)
    MAIN.overview = Overview(MAIN)
    MAIN.spectrum = Spectrum(MAIN)
    MAIN.traces = Traces(MAIN)
    MAIN.video = Video(MAIN)
    MAIN.settings = Settings(MAIN)  # depends on all widgets apart from Info
    MAIN.info = Info(MAIN)  # this has to be the last, it depends on settings

    MAIN.setCentralWidget(MAIN.traces)

    """ ------ LIST DOCKWIDGETS ------ """
    new_docks = [{'name': 'Information',
                  'widget': MAIN.info,
                  'main_area': Qt.LeftDockWidgetArea,
                  'extra_area': Qt.RightDockWidgetArea,
                  },
                 {'name': 'Labels',
                  'widget': MAIN.labels,
                  'main_area': Qt.RightDockWidgetArea,
                  'extra_area': Qt.LeftDockWidgetArea,
                  },
                 {'name': 'Channels',
                  'widget': MAIN.channels,
                  'main_area': Qt.RightDockWidgetArea,
                  'extra_area': Qt.LeftDockWidgetArea,
                  },
                 {'name': 'Spectrum',
                  'widget': MAIN.spectrum,
                  'main_area': Qt.RightDockWidgetArea,
                  'extra_area': Qt.LeftDockWidgetArea,
                  },
                 {'name': 'Annotations',
                  'widget': MAIN.notes,
                  'main_area': Qt.LeftDockWidgetArea,
                  'extra_area': Qt.RightDockWidgetArea,
                  },
                 {'name': 'Video',
                  'widget': MAIN.video,
                  'main_area': Qt.LeftDockWidgetArea,
                  'extra_area': Qt.RightDockWidgetArea,
                  },
                 {'name': 'Overview',
                  'widget': MAIN.overview,
                  'main_area': Qt.BottomDockWidgetArea,
                  'extra_area': Qt.TopDockWidgetArea,
                  },
                 ]

    """ ------ CREATE DOCKWIDGETS ------ """
    idx_docks = {}
    actions = MAIN.action

    actions['dockwidgets'] = []
    for dock in new_docks:
        dockwidget = QDockWidget(dock['name'], MAIN)
        dockwidget.setWidget(dock['widget'])
        dockwidget.setAllowedAreas(dock['main_area'] | dock['extra_area'])
        dockwidget.setObjectName(dock['name'])  # savestate

        idx_docks[dock['name']] = dockwidget
        MAIN.addDockWidget(dock['main_area'], dockwidget)

        dockwidget_action = dockwidget.toggleViewAction()
        dockwidget_action.setIcon(QIcon(ICON['widget']))

        actions['dockwidgets'].append(dockwidget_action)

    """ ------ ORGANIZE DOCKWIDGETS ------ """
    MAIN.tabifyDockWidget(idx_docks['Information'],
                          idx_docks['Video'])
    MAIN.tabifyDockWidget(idx_docks['Channels'],
                          idx_docks['Labels'])
    idx_docks['Information'].raise_()


def create_actions(MAIN):
    """Create all the possible actions."""
    actions = MAIN.action  # actions was already taken

    """ ------ OPEN SETTINGS ------ """
    actions['open_settings'] = QAction(QIcon(ICON['settings']), 'Settings',
                                       MAIN)
    actions['open_settings'].triggered.connect(MAIN.show_settings)

    """ ------ CLOSE WINDOW ------ """
    actions['close_wndw'] = QAction(QIcon(ICON['quit']), 'Quit', MAIN)
    actions['close_wndw'].triggered.connect(MAIN.close)

    """ ------ ABOUT ------ """
    actions['about'] = QAction('About WONAMBI', MAIN)
    actions['about'].triggered.connect(MAIN.about)

    actions['aboutqt'] = QAction('About Qt', MAIN)
    actions['aboutqt'].triggered.connect(lambda: QMessageBox.aboutQt(MAIN))


def create_menubar(MAIN):
    """Create the whole menubar, based on actions."""
    actions = MAIN.action
    menubar = MAIN.menuBar()
    menubar.clear()

    """ ------ FILE ------ """
    menu_file = menubar.addMenu('File')
    menu_file.addAction(MAIN.info.action['open_dataset'])
    submenu_recent = menu_file.addMenu('Recent Datasets')
    submenu_recent.addActions(MAIN.info.action['open_recent'])
    menu_file.addAction(MAIN.info.action['export'])

    menu_file.addSeparator()
    menu_file.addAction(actions['open_settings'])
    menu_file.addSeparator()
    menu_file.addAction(actions['close_wndw'])

    """ ------ CHANNELS ------ """
    actions = MAIN.channels.action
    menu_time = menubar.addMenu('Channels')
    menu_time.addAction(actions['load_channels'])
    menu_time.addAction(actions['save_channels'])

    """ ------ NAVIGATION ------ """
    actions = MAIN.traces.action

    menu_time = menubar.addMenu('Navigation')
    menu_time.addAction(actions['step_prev'])
    menu_time.addAction(actions['step_next'])
    menu_time.addAction(actions['page_prev'])
    menu_time.addAction(actions['page_next'])
    menu_time.addSeparator()
    menu_time.addAction(actions['addtime_-6h'])
    menu_time.addAction(actions['addtime_-1h'])
    menu_time.addAction(actions['addtime_-10min'])
    menu_time.addAction(actions['addtime_10min'])
    menu_time.addAction(actions['addtime_1h'])
    menu_time.addAction(actions['addtime_6h'])
    menu_time.addSeparator()
    menu_time.addAction(actions['next_event'])
    menu_time.addAction(actions['del_and_next_event'])
    menu_time.addAction(actions['next_of_same_type'])
    menu_time.addAction(actions['centre_event'])
    menu_time.addSeparator()
    menu_time.addAction(actions['go_to_epoch'])
    menu_time.addAction(actions['line_up_with_epoch'])

    """ ------ VIEW ------ """
    actions = MAIN.traces.action

    menu_view = menubar.addMenu('View')
    submenu_ampl = menu_view.addMenu('Global Scaling')
    submenu_ampl.addAction(actions['Y_less'])
    submenu_ampl.addAction(actions['Y_more'])
    submenu_ampl.addSeparator()
    for x in sorted(MAIN.value('y_scale_presets'), reverse=True):
        act = submenu_ampl.addAction('Set to ' + str(x))
        act.triggered.connect(partial(MAIN.traces.Y_ampl, x))

    submenu_dist = menu_view.addMenu('Distance Between Traces')
    submenu_dist.addAction(actions['Y_wider'])
    submenu_dist.addAction(actions['Y_tighter'])
    submenu_dist.addSeparator()
    for x in sorted(MAIN.value('y_distance_presets'),
                    reverse=True):
        act = submenu_dist.addAction('Set to ' + str(x))
        act.triggered.connect(partial(MAIN.traces.Y_dist, x))

    submenu_length = menu_view.addMenu('Window Length')
    submenu_length.addAction(actions['X_more'])
    submenu_length.addAction(actions['X_less'])
    submenu_length.addSeparator()
    for x in sorted(MAIN.value('window_length_presets'),
                    reverse=True):
        act = submenu_length.addAction('Set to ' + str(x))
        act.triggered.connect(partial(MAIN.traces.X_length, x))

    menu_view.addAction(actions['cross_chan_mrk'])

    menu_view.addSeparator()
    menu_view.addAction(actions['export_svg'])

    """ ------ ANNOTATIONS ------ """
    actions = MAIN.notes.action

    menu_annot = menubar.addMenu('Annotations')
    menu_annot.addAction(actions['new_annot'])
    menu_annot.addAction(actions['load_annot'])
    menu_annot.addAction(actions['clear_annot'])
    menu_annot.addSeparator()

    submenu_rater = menu_annot.addMenu('Rater')
    submenu_rater.addAction(actions['new_rater'])
    submenu_rater.addAction(actions['del_rater'])
    submenu_rater.addSeparator()
    if MAIN.notes.annot is not None:
        for rater in sorted(MAIN.notes.annot.raters):
            act = submenu_rater.addAction(rater)
            act.triggered.connect(partial(MAIN.notes.select_rater, rater))
    menu_annot.addSeparator()

    submenu_marker = menu_annot.addMenu('Bookmark')
    submenu_marker.addAction(actions['new_bookmark'])

    submenu_event = menu_annot.addMenu('Event')
    submenu_event.addAction(actions['new_eventtype'])
    submenu_event.addAction(actions['del_eventtype'])
    submenu_event.addAction(actions['rename_eventtype'])
    submenu_mrk2evt = submenu_event.addMenu('Markers into events')
    submenu_mrk2evt.addAction(actions['m2e_newname'])
    submenu_mrk2evt.addAction(actions['m2e_keepname'])
    submenu_event.addAction(actions['merge_events'])
    submenu_event.addAction(MAIN.traces.action['change_event_type'])

    # these are the real QActions attached to notes
    submenu_stage = menu_annot.addMenu('Stage')
    submenu_stage.addActions(MAIN.notes.actions())

    submenu_mrkr = menu_annot.addMenu('Cycle')
    submenu_mrkr.addAction(actions['cyc_start'])
    submenu_mrkr.addAction(actions['cyc_end'])
    submenu_mrkr.addAction(actions['remove_cyc'])
    submenu_mrkr.addAction(actions['clear_cyc'])
    menu_annot.addSeparator()

    submenu_import = menu_annot.addMenu('Import staging')
    submenu_import.addAction(actions['import_alice'])
    submenu_import.addAction(actions['import_compumedics'])
    submenu_import.addAction(actions['import_domino'])
    submenu_import.addAction(actions['import_prana'])
    submenu_import.addAction(actions['import_remlogic'])
    submenu_import.addAction(actions['import_sandman'])
    submenu_import.addAction(actions['import_fasst'])

    submenu_import_qual = menu_annot.addMenu('Import signal quality')
    submenu_import_qual.addAction(actions['import_alice_qual'])
    submenu_import_qual.addAction(actions['import_compumedics_qual'])
    submenu_import_qual.addAction(actions['import_domino_qual'])
    submenu_import_qual.addAction(actions['import_prana_qual'])
    submenu_import_qual.addAction(actions['import_remlogic_qual'])
    submenu_import_qual.addAction(actions['import_sandman_qual'])
    
    menu_annot.addAction(actions['imp_evt_csv'])
    menu_annot.addSeparator()
    
    submenu_export_staging = menu_annot.addMenu('Export staging')
    submenu_export_staging.addAction(actions['export_to_csv'])
    submenu_export_staging.addAction(actions['export_to_remlogic'])
    submenu_export_staging.addAction(actions['export_to_remlogic_fr'])

    menu_annot.addAction(actions['export_events'])
    
#==============================================================================
#     submenu_exp_evt = menu_annot.addMenu('Export events')
#     submenu_exp_evt.addAction(actions['exp_evt_csv'])
#     submenu_exp_evt.addAction(actions['exp_evt_brainvision'])
#==============================================================================
    
    menu_annot.addSeparator()
    
    menu_annot.addAction(actions['export_sleepstats'])

    """ ------ ANALYSIS ------ """
    actions = MAIN.notes.action

    menu_analysis = menubar.addMenu('Analysis')

    submenu_detect = menu_analysis.addMenu('Detection')
    submenu_detect.addAction(actions['spindle'])
    submenu_detect.addAction(actions['slow_wave'])

    menu_analysis.addAction(actions['analyze'])


    """ ------ WINDOWS ------ """
    actions = MAIN.action

    menu_window = menubar.addMenu('Windows')
    for dockwidget_act in actions['dockwidgets']:
        menu_window.addAction(dockwidget_act)
    MAIN.menu_window = menu_window

    menu_about = menubar.addMenu('About')
    menu_about.addAction(actions['about'])
    menu_about.addAction(actions['aboutqt'])

def create_toolbar(MAIN):
    """Create the various toolbars."""
    actions = MAIN.action

    toolbar = MAIN.addToolBar('File Management')
    toolbar.setObjectName('File Management')  # for savestate
    toolbar.addAction(MAIN.info.action['open_dataset'])
    toolbar.addSeparator()
    toolbar.addAction(MAIN.channels.action['load_channels'])
    toolbar.addAction(MAIN.channels.action['save_channels'])
    toolbar.addSeparator()
    toolbar.addAction(MAIN.notes.action['new_annot'])
    toolbar.addAction(MAIN.notes.action['load_annot'])

    """ ------ SCROLL ------ """
    actions = MAIN.traces.action

    toolbar = MAIN.addToolBar('Scroll')
    toolbar.setObjectName('Scroll')  # for savestate
    toolbar.addAction(actions['step_prev'])
    toolbar.addAction(actions['step_next'])
    toolbar.addAction(actions['page_prev'])
    toolbar.addAction(actions['page_next'])
    toolbar.addSeparator()
    toolbar.addAction(actions['X_more'])
    toolbar.addAction(actions['X_less'])
    toolbar.addSeparator()
    toolbar.addAction(actions['Y_less'])
    toolbar.addAction(actions['Y_more'])
    toolbar.addAction(actions['Y_wider'])
    toolbar.addAction(actions['Y_tighter'])

    """ ------ ANNOTATIONS ------ """
    actions = MAIN.notes.action

    toolbar = MAIN.addToolBar('Annotations')
    toolbar.setObjectName('Annotations')

    toolbar.addAction(actions['new_bookmark'])
    toolbar.addSeparator()
    toolbar.addAction(actions['new_event'])
    toolbar.addWidget(MAIN.notes.idx_eventtype)
    toolbar.addSeparator()
    toolbar.addWidget(MAIN.notes.idx_stage)
    toolbar.addWidget(MAIN.notes.idx_quality)



from pytest import approx, raises

from wonambi import Dataset
from wonambi.detect.spindle import DetectSpindle

from .paths import psg_file

d = Dataset(psg_file)
data = d.read_data(chan=('EEG Fpz-Cz', 'EEG Pz-Oz'), begtime=35790, endtime=35820)


def test_detect_spindle_Moelle2011():
    detsp = DetectSpindle()
    assert repr(detsp) == 'detsp_Moelle2011_09-15Hz_00.5-03.0s'

    sp = detsp(data)
    assert len(sp.events) == 2


def test_detect_spindle_Nir2011():
    detsp = DetectSpindle(method='Nir2011')

    sp = detsp(data)
    assert len(sp.events) == 2


def test_detect_spindle_Wamsley2012():
    detsp = DetectSpindle(method='Wamsley2012')

    sp = detsp(data)
    assert len(sp.events) == 1


def test_detect_spindle_Ferrarelli2007():
    detsp = DetectSpindle(method='Ferrarelli2007')

    sp = detsp(data)
    assert len(sp.events) == 0


def test_detect_Ray2015():
    detsp = DetectSpindle(method='Ray2015')

    sp = detsp(data)
    assert len(sp.events) == 3
    
    
def test_detect_Lacourse2018():
    detsp = DetectSpindle(method='Lacourse2018')

    sp = detsp(data)
    assert len(sp.events) == 4

def test_detect_spindle_FASST():
    detsp = DetectSpindle(method='FASST')

    sp = detsp(data)
    assert len(sp.events) == 4

def test_detect_spindle_FASST2():
    detsp = DetectSpindle(method='FASST2')

    sp = detsp(data)
    assert len(sp.events) == 6

def test_detect_spindle_UCSD():
    detsp = DetectSpindle(method='UCSD')

    sp = detsp(data)
    assert len(sp.events) == 6

def test_detect_spindle_Concordia():
    detsp = DetectSpindle(method='Concordia')

    sp = detsp(data)
    assert (len(sp.events)) == 2

def test_detect_spindle_unknownmethod():
    with raises(ValueError):
        detsp = DetectSpindle(method='xxx')

def test_merge_close():
    detsp = DetectSpindle(method='Nir2011')

    sp = detsp(data)
    assert len(sp.events) == 2

def test_detect_spindle_to_data():
    detsp = DetectSpindle()
    sp = detsp(data)

    sp_data = sp.to_data('count')
    assert sp_data(0)[0] == 1

    sp_freq = sp.to_data('peak_freq')
    assert approx(sp_freq(0)[0]) == 13.793103448275861



"""Widget to redefine the labels.
"""
from logging import getLogger
from pathlib import Path
from re import split

from PyQt5.QtCore import Qt
from PyQt5.QtGui import (QColor,
                         )

from PyQt5.QtWidgets import (QAbstractItemView,
                             QFileDialog,
                             QHBoxLayout,
                             QPushButton,
                             QVBoxLayout,
                             QTableWidget,
                             QTableWidgetItem,
                             QWidget
                             )

lg = getLogger(__name__)


class Labels(QWidget):
    """
    Attributes
    ----------
    chan_name : list of str
        list of all the labels (with the user-defined changes)
    """
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.filename = None
        self.chan_name = None  # None when dataset is not loaded

        self.create()

    def create(self):

        self.idx_load = QPushButton('Load')
        self.idx_load.clicked.connect(self.load_labels)
        self.idx_load.setToolTip('Load file with a list of channels (separated by , or ; or tabs or spaces).')
        self.idx_save = QPushButton('Save')
        self.idx_save.clicked.connect(self.save_labels)
        self.idx_save.setEnabled(False)

        # cancel is equal to setting labels to what they were
        self.idx_cancel = QPushButton('Cancel')
        self.idx_cancel.clicked.connect(self.update)
        self.idx_apply = QPushButton('Apply')
        self.idx_apply.clicked.connect(self.apply)
        self.idx_apply.setToolTip('Changes will take effect. This will reset the channel groups and traces.')

        layout_0 = QHBoxLayout()
        layout_0.addWidget(self.idx_load)
        layout_0.addWidget(self.idx_save)

        layout_1 = QHBoxLayout()
        layout_1.addWidget(self.idx_cancel)
        layout_1.addWidget(self.idx_apply)

        self.table = QTableWidget()

        self.table.horizontalHeader().setStretchLastSection(True)
        self.table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table.setColumnCount(2)
        self.table.setHorizontalHeaderLabels(['Current Labels', 'New Labels'])

        self.table.cellChanged.connect(self.check_labels)

        layout = QVBoxLayout()
        layout.addLayout(layout_0)
        layout.addWidget(self.table)
        layout.addLayout(layout_1)

        self.setLayout(layout)

        self.setEnabled(False)

    def update(self, checked=False, labels=None, custom_labels=None):
        """Use this function when we make changes to the list of labels or when
        we load a new dataset.

        Parameters
        ----------
        checked : bool
            argument from clicked.connect
        labels : list of str
            list of labels in the dataset (default)
        custom_labels : list of str
            list of labels from a file
        """
        if labels is not None:
            self.setEnabled(True)
            self.chan_name = labels

        self.table.blockSignals(True)
        self.table.clearContents()
        self.table.setRowCount(len(self.chan_name))

        for i, label in enumerate(self.chan_name):
            old_label = QTableWidgetItem(label)
            old_label.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)

            if custom_labels is not None and i < len(custom_labels) and custom_labels[i]:  # it's not empty string or None
                label_txt = custom_labels[i]
            else:
                label_txt = label
            new_label = QTableWidgetItem(label_txt)

            self.table.setItem(i, 0, old_label)
            self.table.setItem(i, 1, new_label)

        self.table.blockSignals(False)

    def check_labels(self):

        # read new labels first
        labels = self._read_labels()

        # disable apply, if there are duplicates
        if len(labels) == len(set(labels)):
            self.idx_apply.setEnabled(True)
        else:
            self.idx_apply.setEnabled(False)

        # mark duplicates in red
        self.table.blockSignals(True)
        for i, label in enumerate(labels):
            if labels.count(label) > 1:
                self.table.item(i, 1).setBackground(QColor('red'))
            else:
                self.table.item(i, 1).setBackground(QColor('white'))
        self.table.blockSignals(False)

    def load_labels(self, checked=False, test_name=None):
        if self.filename is not None:
            filename = self.filename
        elif self.parent.info.filename is not None:
            filename = Path(self.parent.info.filename)
        else:
            filename = None

        if test_name:
            filename = test_name
        else:
            filename, _ = QFileDialog.getOpenFileName(self,
                                                      'Open Labels',
                                                      str(filename.parent),
                                                      'Comma-separated values (*.csv);; Text file (*.txt);; All Files(*.*)')
        if filename == '':
            return

        self.filename = Path(filename)

        with self.filename.open() as f:
            text = f.read()

        labels = split(', |,|; |;|\t|\n| ', text)
        labels = [label.strip() for label in labels]
        self.update(custom_labels=labels)

    def save_labels(self):
        """Save labels modified by the user.

        TODO
        ----
        Save labels modified by the user
        """
        pass

    def apply(self):

        self.chan_name = self._read_labels()
        self.parent.info.dataset.header['chan_name'] = self.chan_name

        self.parent.channels.reset()
        self.parent.channels.update()
        self.parent.traces.reset()

    def reset(self):
        self.table.blockSignals(True)
        self.table.clearContents()
        self.table.blockSignals(False)

        self.setEnabled(False)

    def _read_labels(self):

        labels = []
        for i in range(self.table.rowCount()):
            labels.append(self.table.item(i, 1).text())

        return labels



"""Module to select periods of interest, based on number of trials or any of
the axes.

There is some overlap between Select and the Data.__call__(). The main
difference is that Select takes an instance of Data as input and returns
another instance of Data as output, whil Data.__call__() returns the actual
content of the data.

Select should be as flexible as possible. There are quite a few cases, which
will be added as we need them.
"""
from collections import Iterable
from logging import getLogger

from numpy import (arange, asarray, diff, empty, hstack, inf, linspace,
                   nan_to_num, ones, ravel, setdiff1d)
from numpy.lib.stride_tricks import as_strided
from math import isclose
from scipy.signal import decimate

try:
    from PyQt5.QtCore import Qt
    from PyQt5.QtWidgets import QProgressDialog
except ImportError:
    Qt = None
    QProgressDialog = None

from .. import ChanTime
from .montage import montage
from .reject import remove_artf_evts

lg = getLogger(__name__)


class Segments():
    """Class containing a set of data segments for analysis, with metadata.
    Only contains metadata until .read_data is called.
    
    Attributes
    ----------
    dataset : instance of wonambi.dataset
        metadata for the associated record
    segments : list of dict
        chronological list of segment metadata. Each segment dict contains info
        about start and end times, stage, cycle, channel and event name, if 
        applicable. Once read_data is called, the signal data are added to each
        segment dictionary under 'data'.
    """
    def __init__(self, dataset):
        self.dataset = dataset
        self.segments = []

    def __iter__(self):
        for one_event in self.segments:
            yield one_event

    def __len__(self):
        return len(self.segments)

    def __getitem__(self, index):
        return self.segments[index]

    def read_data(self, chan=[], ref_chan=[], grp_name=None, concat_chan=False, 
                  max_s_freq=30000, parent=None):
        """Read data for analysis. Adds data as 'data' in each dict.

        Parameters
        ----------
        chan : list of str
            active channel names as they appear in record, without ref or group
        ref_chan : list of str
            reference channel names as they appear in record, without group
        grp_name : str
            name of the channel group, required in GUI
        concat_chan : bool
            if True, data from all channels will be concatenated
        max_s_freq: : int
            maximum sampling frequency
        parent : QWidget
            for GUI only. Identifies parent widget for display of progress
            dialog.
        """
        output = []

        # Set up Progress Bar
        if parent:
            n_subseg = sum([len(x['times']) for x in self.segments])
            progress = QProgressDialog('Fetching signal', 'Abort', 0, n_subseg,
                                       parent)
            progress.setWindowModality(Qt.ApplicationModal)
            counter = 0

        # Begin bundle loop; will yield one segment per loop
        for i, seg in enumerate(self.segments):
            one_segment = ChanTime()
            one_segment.axis['chan'] = empty(1, dtype='O')
            one_segment.axis['time'] = empty(1, dtype='O')
            one_segment.data = empty(1, dtype='O')
            subseg = []

            # Subsegment loop; subsegments will be concatenated
            for t0, t1 in seg['times']:
                if parent:
                    progress.setValue(counter)
                    counter += 1

                # if channel not specified, use segment channel
                if chan:
                    active_chan = chan
                elif seg['chan']:
                    active_chan = [seg['chan'].split(' (')[0]]
                else:
                    raise ValueError('No channel was specified and the '
                                     'segment at {}-{} has no channel.'.format(
                                             t0, t1))
                active_chan = chan if chan else [seg['chan'].split(' (')[0]]
                chan_to_read = active_chan + ref_chan

                data = self.dataset.read_data(chan=chan_to_read, begtime=t0,
                                         endtime=t1)

                # Downsample if necessary
                if data.s_freq > max_s_freq:
                    q = int(data.s_freq / max_s_freq)
                    lg.debug('Decimate (no low-pass filter) at ' + str(q))

                    data.data[0] = data.data[0][:, slice(None, None, q)]
                    data.axis['time'][0] = data.axis['time'][0][slice(
                            None, None, q)]
                    data.s_freq = int(data.s_freq / q)

                # read data from disk
                subseg.append(_create_data(
                    data, active_chan, ref_chan=ref_chan, grp_name=grp_name))

            one_segment.s_freq = s_freq = data.s_freq
            one_segment.axis['chan'][0] = chs = subseg[0].axis['chan'][0]
            one_segment.axis['time'][0] = timeline = hstack(
                    [x.axis['time'][0] for x in subseg])
            one_segment.data[0] = empty((len(active_chan), len(timeline)),
                                        dtype='f')
            n_stitch = sum(asarray(diff(timeline) > 2/s_freq, dtype=bool))

            for i, ch in enumerate(subseg[0].axis['chan'][0]):
                    one_segment.data[0][i, :] = hstack(
                            [x(chan=ch)[0] for x in subseg])

            # For channel concatenation
            if concat_chan and len(chs) > 1:
                one_segment.data[0] = ravel(one_segment.data[0])
                one_segment.axis['chan'][0] = asarray([(', ').join(chs)],
                                dtype='U')
                # axis['time'] should not be used in this case

            output.append({'data': one_segment,
                           'chan': active_chan,
                           'stage': seg['stage'],
                           'cycle': seg['cycle'],
                           'name': seg['name'],
                           'n_stitch': n_stitch
                           })

            if parent:
                if progress.wasCanceled():
                    parent.parent.statusBar().showMessage('Process canceled by'
                                           ' user.')
                    return

        if parent:
            progress.setValue(counter)

        self.segments = output

        return 1 # for GUI


def select(data, trial=None, invert=False, **axes_to_select):
    """Define the selection of trials, using ranges or actual values.

    Parameters
    ----------
    data : instance of Data
        data to select from.
    trial : list of int or ndarray (dtype='i'), optional
        index of trials of interest
    **axes_to_select, optional
        Values need to be tuple or list. If the values in one axis are string,
        then you need to specify all the strings that you want. If the values
        are numeric, then you should specify the range (you cannot specify
        single values, nor multiple values). To select only up to one point,
        you can use (None, value_of_interest)
    invert : bool
        take the opposite selection

    Returns
    -------
    instance, same class as input
        data where selection has been applied.
    """
    if trial is not None and not isinstance(trial, Iterable):
        raise TypeError('Trial needs to be iterable.')

    for axis_to_select, values_to_select in axes_to_select.items():
        if (not isinstance(values_to_select, Iterable) or
           isinstance(values_to_select, str)):
            raise TypeError(axis_to_select + ' needs to be iterable.')

    if trial is None:
        trial = range(data.number_of('trial'))
    else:
        trial = trial
        if invert:
            trial = setdiff1d(range(data.number_of('trial')), trial)

    # create empty axis
    output = data._copy(axis=False)
    for one_axis in output.axis:
        output.axis[one_axis] = empty(len(trial), dtype='O')
    output.data = empty(len(trial), dtype='O')

    to_select = {}
    for cnt, i in enumerate(trial):
        lg.debug('Selection on trial {0: 6}'.format(i))
        for one_axis in output.axis:
            values = data.axis[one_axis][i]

            if one_axis in axes_to_select.keys():
                values_to_select = axes_to_select[one_axis]

                if len(values_to_select) == 0:
                    selected_values = ()

                elif isinstance(values_to_select[0], str):
                    selected_values = asarray(values_to_select, dtype='U')

                else:
                    if (values_to_select[0] is None and
                       values_to_select[1] is None):
                        bool_values = ones(len(values), dtype=bool)
                    elif values_to_select[0] is None:
                        bool_values = values < values_to_select[1]
                    elif values_to_select[1] is None:
                        bool_values = values_to_select[0] <= values
                    else:
                        bool_values = ((values_to_select[0] <= values) &
                                       (values < values_to_select[1]))
                    selected_values = values[bool_values]

                if invert:
                    selected_values = setdiff1d(values, selected_values)

                lg.debug('In axis {0}, selecting {1: 6} '
                         'values'.format(one_axis,
                                         len(selected_values)))
                to_select[one_axis] = selected_values
            else:
                lg.debug('In axis ' + one_axis + ', selecting all the '
                         'values')
                selected_values = data.axis[one_axis][i]

            output.axis[one_axis][cnt] = selected_values

        output.data[cnt] = data(trial=i, **to_select)

    return output


def resample(data, s_freq=None, axis='time', ftype='fir', n=None):
    """Downsample the data after applying a filter.

    Parameters
    ----------
    data : instance of Data
        data to downsample
    s_freq : int or float
        desired sampling frequency
    axis : str
        axis you want to apply downsample on (most likely 'time')
    ftype : str
        filter type to apply. The default here is 'fir', like Matlab but unlike
        the default in scipy, because it works better
    n : int
        The order of the filter (1 less than the length for ‘fir’).

    Returns
    -------
    instance of Data
        downsampled data
    """
    output = data._copy()
    ratio = int(data.s_freq / s_freq)

    for i in range(data.number_of('trial')):
        output.data[i] = decimate(data.data[i], ratio,
                                  axis=data.index_of(axis),
                                  zero_phase=True)

        n_samples = output.data[i].shape[data.index_of(axis)]
        output.axis[axis][i] = linspace(data.axis[axis][i][0],
                                        data.axis[axis][i][-1] +
                                        1 / data.s_freq,
                                        n_samples)

    output.s_freq = s_freq

    return output


def fetch(dataset, annot, cat=(0, 0, 0, 0), evt_type=None, stage=None,
          cycle=None, chan_full=None, epoch=None, epoch_dur=30,
          epoch_overlap=0, epoch_step=None, reject_epoch=False,
          reject_artf=False, min_dur=0):
    """Create instance of Segments for analysis, complete with info about
    stage, cycle, channel, event type. Segments contains only metadata until
    .read_data is called.

    Parameters
    ----------
    dataset : instance of Dataset
        info about record
    annot : instance of Annotations
        scoring info
    cat : tuple of int
        Determines where the signal is concatenated.
        If cat[0] is 1, cycles will be concatenated.
        If cat[1] is 1, different stages will be concatenated.
        If cat[2] is 1, discontinuous signal within a same condition
        (stage, cycle, event type) will be concatenated.
        If cat[3] is 1, events of different types will be concatenated.
        0 in any position indicates no concatenation.
    evt_type: list of str, optional
        Enter a list of event types to get events; otherwise, epochs will
        be returned.
    stage: list of str, optional
        Stage(s) of interest. If None, stage is ignored.
    cycle: list of tuple of two float, optional
        Cycle(s) of interest, as start and end times in seconds from record
        start. If None, cycles are ignored.
    chan_full: list of str or None
        Channel(s) of interest, only used for events (epochs have no
        channel). Channel format is 'chan_name (group_name)'.
        If used for epochs, separate segments will be returned for each 
        channel; this is necessary for channel-specific artefact removal (see
        reject_artf below). If None, channel is ignored.
    epoch : str, optional
        If 'locked', returns epochs locked to staging. If 'unlocked', divides
        signal (with specified concatenation) into epochs of duration epoch_dur
        starting at first sample of every segment and discarding any remainder.
        If None, longest run of signal is returned.
    epoch_dur : float
        only for epoch='unlocked'. Duration of epochs returned, in seconds.
    epoch_overlap : float
        only for epoch='unlocked'. Ratio of overlap between two consecutive
        segments. Value between 0 and 1. Overriden by step.
    epoch_step : float
        only for epoch='unlocked'. Time between consecutive epoch starts, in
        seconds. Overrides epoch_overlap/
    reject_epoch: bool
        If True, epochs marked as 'Poor' quality or staged as 'Artefact' will
        be rejected (and the signal segmented in consequence). Has no effect on
        event selection.
    reject_artf : bool
        If True, excludes events marked as 'Artefact'. If chan_full is 
        specified, only artefacts marked on a given channel are removed from 
        that channel. Signal is segmented in consequence. 
        If None, Artefact events are ignored.
    min_dur : float
        Minimum duration of segments returned, in seconds.

    Returns
    -------
    instance of Segments
        metadata for all analysis segments
    """
    bundles = get_times(annot, evt_type=evt_type, stage=stage, cycle=cycle,
                        chan=chan_full, exclude=reject_epoch)

    # Remove artefacts
    if reject_artf and bundles:
        for bund in bundles:
            bund['times'] = remove_artf_evts(bund['times'], annot, 
                bund['chan'], min_dur=0)

    # Divide bundles into segments to be concatenated
    if bundles:

        if 'locked' == epoch:
            bundles = _divide_bundles(bundles)

        elif 'unlocked' == epoch:

            if epoch_step is not None:
                step = epoch_step
            else:
                step = epoch_dur - (epoch_dur * epoch_overlap)

            bundles = _concat(bundles, cat)
            bundles = _find_intervals(bundles, epoch_dur, step)

        elif not epoch:
            bundles = _concat(bundles, cat)

        # Minimum duration
        bundles = _longer_than(bundles, min_dur)
    
    segments = Segments(dataset)
    segments.segments = bundles

    return segments


def get_times(annot, evt_type=None, stage=None, cycle=None, chan=None,
              exclude=False):
    """Get start and end times for selected segments of data, bundled
    together with info.

    Parameters
    ----------
    annot: instance of Annotations
        The annotation file containing events and epochs
    evt_type: list of str, optional
        Enter a list of event types to get events; otherwise, epochs will
        be returned.
    stage: list of str, optional
        Stage(s) of interest. If None, stage is ignored.
    cycle: list of tuple of two float, optional
        Cycle(s) of interest, as start and end times in seconds from record
        start. If None, cycles are ignored.
    chan: list of str or tuple of None
        Channel(s) of interest. Channel format is 'chan_name (group_name)'.
        If None, channel is ignored.
    exclude: bool
        Exclude epochs by quality. If True, epochs marked as 'Poor' quality
        or staged as 'Artefact' will be rejected (and the signal segmented
        in consequence). Has no effect on event selection.

    Returns
    -------
    list of dict
        Each dict has times (the start and end times of each segment, as
        list of tuple of float), stage, cycle, chan, name (event type,
        if applicable)

    Notes
    -----
    This function returns epoch or event start and end times, bundled
    together according to the specified parameters.
    Presently, setting exclude to True does not exclude events found in Poor
    signal epochs. The rationale is that events would never be marked in Poor
    signal epochs. If they were automatically detected, these epochs would
    have been left out during detection. If they were manually marked, then
    it must have been Good signal. At the moment, in the GUI, the exclude epoch
    option is disabled when analyzing events, but we could fix the code if we
    find a use case for rejecting events based on the quality of the epoch
    signal.
    """
    getter = annot.get_epochs

    if stage is None:
        stage = (None,)
    if cycle is None:
        cycle = (None,)
    if chan is None:
        chan = (None,)
    if evt_type is None:
        evt_type = (None,)
    elif isinstance(evt_type[0], str):
        getter = annot.get_events
        if chan != (None,):
            chan.append('') # also retrieve events marked on all channels
    else:
        lg.error('Event type must be list/tuple of str or None')

    qual = None
    if exclude:
        qual = 'Good'

    bundles = []
    for et in evt_type:

        for ch in chan:

            for cyc in cycle:

                for ss in stage:

                    st_input = ss
                    if ss is not None:
                        st_input = (ss,)

                    evochs = getter(name=et, time=cyc, chan=(ch,),
                                    stage=st_input, qual=qual)
                    if evochs:
                        times = [(e['start'], e['end']) for e in evochs]
                        times = sorted(times, key=lambda x: x[0])
                        one_bundle = {'times': times,
                                      'stage': ss,
                                      'cycle': cyc,
                                      'chan': ch,
                                      'name': et}
                        bundles.append(one_bundle)

    return bundles


def _longer_than(segments, min_dur):
    """Remove segments longer than min_dur."""
    if min_dur <= 0.:
        return segments

    long_enough = []
    for seg in segments:

        if sum([t[1] - t[0] for t in seg['times']]) >= min_dur:
            long_enough.append(seg)

    return long_enough


def _concat(bundles, cat=(0, 0, 0, 0)):
    """Prepare event or epoch start and end times for concatenation."""
    chan = sorted(set([x['chan'] for x in bundles]))
    cycle = sorted(set([x['cycle'] for x in bundles]))
    stage = sorted(set([x['stage'] for x in bundles]))
    evt_type = sorted(set([x['name'] for x in bundles]))

    all_cycle = None
    all_stage = None
    all_evt_type = None

    if cycle[0] is not None:
        all_cycle = ', '.join([str(c) for c in cycle])
    if stage[0] is not None:
        all_stage = ', '.join(stage)
    if evt_type[0] is not None:
        all_evt_type = ', '.join(evt_type)

    if cat[0]:
        cycle = [all_cycle]

    if cat[1]:
        stage = [all_stage]

    if cat[3]:
        evt_type = [all_evt_type]

    to_concat = []
    for ch in chan:

        for cyc in cycle:

            for st in stage:

                for et in evt_type:
                    new_times = []

                    for bund in bundles:
                        chan_cond = ch == bund['chan']
                        cyc_cond = cyc in (bund['cycle'], all_cycle)
                        st_cond = st in (bund['stage'], all_stage)
                        et_cond = et in (bund['name'], all_evt_type)

                        if chan_cond and cyc_cond and st_cond and et_cond:
                            new_times.extend(bund['times'])

                    new_times = sorted(new_times, key=lambda x: x[0])
                    new_bund = {'times': new_times,
                              'chan': ch,
                              'cycle': cyc,
                              'stage': st,
                              'name': et
                              }
                    to_concat.append(new_bund)

    if not cat[2]:
        to_concat_new = []

        for bund in to_concat:
            last = None
            bund['times'].append((inf,inf))
            start = 0

            for i, j in enumerate(bund['times']):

                if last is not None:
                    if not isclose(j[0], last, abs_tol=0.01):
                        new_times = bund['times'][start:i]
                        new_bund = bund.copy()
                        new_bund['times'] = new_times
                        to_concat_new.append(new_bund)
                        start = i
                last = j[1]

        to_concat = to_concat_new

    to_concat = [x for x in to_concat if x['times']]

    return to_concat


def _divide_bundles(bundles):
    """Take each subsegment inside a bundle and put it in its own bundle,
    copying the bundle metadata."""
    divided = []

    for bund in bundles:
        for t in bund['times']:
            new_bund = bund.copy()
            new_bund['times'] = [t]
            divided.append(new_bund)

    return divided


def _find_intervals(bundles, duration, step):
    """Divide bundles into segments of a certain duration and a certain step,
    discarding any remainder."""
    segments = []
    for bund in bundles:
        beg, end = bund['times'][0][0], bund['times'][-1][1]

        if end - beg >= duration:
            new_begs = arange(beg, end, step)

            for t in new_begs:
                seg = bund.copy()
                seg['times'] = [(t, t + duration)]
                segments.append(seg)

    return segments


def _create_data(data, active_chan, ref_chan=[], grp_name=None):
    """Create data after montage.

    Parameters
    ----------
    data : instance of ChanTime
        the raw data
    active_chan : list of str
        the channel(s) of interest, without reference or group
    ref_chan : list of str
        reference channel(s), without group
    grp_name : str
        name of channel group, if applicable

    Returns
    -------
    instance of ChanTime
        the re-referenced data
    """
    output = ChanTime()
    output.s_freq = data.s_freq
    output.start_time = data.start_time
    output.axis['time'] = data.axis['time']
    output.axis['chan'] = empty(1, dtype='O')
    output.data = empty(1, dtype='O')
    output.data[0] = empty((len(active_chan), data.number_of('time')[0]),
                           dtype='f')

    sel_data = _select_channels(data, active_chan + ref_chan)
    data1 = montage(sel_data, ref_chan=ref_chan)
    data1.data[0] = nan_to_num(data1.data[0])

    all_chan_grp_name = []

    for i, chan in enumerate(active_chan):
        chan_grp_name = chan
        if grp_name:
            chan_grp_name = chan + ' (' + grp_name + ')'
        all_chan_grp_name.append(chan_grp_name)

        dat = data1(chan=chan, trial=0)
        output.data[0][i, :] = dat

    output.axis['chan'][0] = asarray(all_chan_grp_name, dtype='U')

    return output


def _create_subepochs(x, nperseg, step):
    """Transform the data into a matrix for easy manipulation

    Parameters
    ----------
    x : 1d ndarray
        actual data values
    nperseg : int
        number of samples in each row to create
    step : int
        distance in samples between rows
    Returns
    -------
    2d ndarray
        a view (i.e. doesn't copy data) of the original x, with shape
        determined by nperseg and step. You should use the last dimension
    """
    axis = x.ndim - 1  # last dim
    nsmp = x.shape[axis]
    stride = x.strides[axis]
    noverlap = nperseg - step
    v_shape = *x.shape[:axis], (nsmp - noverlap) // step, nperseg
    v_strides = *x.strides[:axis], stride * step, stride
    v = as_strided(x, shape=v_shape, strides=v_strides,
                   writeable=False)  # much safer
    return v


def _select_channels(data, channels):
    """Select channels.

    Parameters
    ----------
    data : instance of ChanTime
        data with all the channels
    channels : list
        channels of interest

    Returns
    -------
    instance of ChanTime
        data with only channels of interest

    Notes
    -----
    This function does the same as wonambi.trans.select, but it's much faster.
    wonambi.trans.Select needs to flexible for any data type, here we assume
    that we have one trial, and that channel is the first dimension.

    """
    output = data._copy()
    chan_list = list(data.axis['chan'][0])
    idx_chan = [chan_list.index(i_chan) for i_chan in channels]
    output.data[0] = data.data[0][idx_chan, :]
    output.axis['chan'][0] = asarray(channels)

    return output



from wonambi.viz import Viz3
from wonambi.attr import Surf, Channels

from .paths import (surf_path,
                    chan_path,
                    VIZ_PATH,
                    )


def test_plot3d_surf():

    surf = Surf(surf_path)

    v = Viz3()
    v.add_surf(surf)

    v.save(VIZ_PATH / 'viz3_01_surf.png')
    v.close()


def test_plot3d_surf_color():

    surf = Surf(surf_path)

    v = Viz3()
    v.add_surf(surf, color=(.5, .5, 0))

    v.save(VIZ_PATH / 'viz3_02_surf_color.png')
    v.close()


def test_plot3d_surf_value():

    surf = Surf(surf_path)

    v = Viz3()
    v.add_surf(surf, values=surf.vert[:, 2], limits_c=(0, 50))

    v.save(VIZ_PATH / 'viz3_03_surf_values.png')
    v.close()


def test_plot3d_surf_value_colorbar():

    surf = Surf(surf_path)

    v = Viz3()
    v.add_surf(surf, values=surf.vert[:, 2], limits_c=(0, 50), colorbar=True)

    v.save(VIZ_PATH / 'viz3_04_surf_colorbar.png')
    v.close()


def test_plot3d_surf_chan():

    surf = Surf(surf_path)
    channels = Channels(chan_path)

    v = Viz3()
    v.add_chan(channels)
    v.add_surf(surf, alpha=.8)

    v.save(VIZ_PATH / 'viz3_05_surf_chan.png')
    v.close()


def test_plot3d_surf_chan_color_one():

    surf = Surf(surf_path)
    channels = Channels(chan_path)

    v = Viz3()
    v.add_chan(channels, color=(1, 0, 0))
    v.add_surf(surf, alpha=.5)

    v.save(VIZ_PATH / 'viz3_06_surf_chan_color_one.png')
    v.close()


def test_plot3d_surf_chan_color_one_alpha():

    surf = Surf(surf_path)
    channels = Channels(chan_path)

    v = Viz3()
    v.add_chan(channels, color=(1, 0, 0), alpha=0.5)
    v.add_surf(surf, alpha=.5)

    v.save(VIZ_PATH / 'viz3_07_surf_chan_color_one_alpha.png')
    v.close()


def test_plot3d_surf_chan_color_everyother():

    from numpy import tile

    surf = Surf(surf_path)
    channels = Channels(chan_path)

    color_everyother = tile([[1, 0, 0], [0, 0, 1]], (14, 1))

    v = Viz3()
    v.add_chan(channels, color=color_everyother)
    v.add_surf(surf, alpha=.5)

    v.save(VIZ_PATH / 'viz3_08_surf_chan_color_everyother.png')
    v.close()


def test_plot3d_surf_chan_values():

    from numpy import arange

    surf = Surf(surf_path)
    channels = Channels(chan_path)

    v = Viz3()
    v.add_chan(channels, values=arange(channels.n_chan), colorbar=True)
    v.add_surf(surf, alpha=.5)

    v.save(VIZ_PATH / 'viz3_09_surf_chan_values.png')
    v.close()

    # reuse the limits_c
    v.add_chan(channels, values=arange(channels.n_chan))



from pytest import approx

from wonambi import Dataset
from wonambi.detect.slowwave import DetectSlowWave

from .paths import psg_file

d = Dataset(psg_file)
data = d.read_data(chan=('EEG Fpz-Cz', 'EEG Pz-Oz'), begtime=27930, endtime=27960)


def test_detect_slowwave_Massimini2004():
    detsw = DetectSlowWave()
    detsw.invert = True
    assert repr(detsw) == 'detsw_Massimini2004_0.10-4.00Hz'

    sw = detsw(data)
    assert len(sw.events) == 1


def test_detect_slowwave_AASM_Massimini2004():
    detsw = DetectSlowWave(method='AASM/Massimini2004')
    detsw.invert = True
    assert repr(detsw) == 'detsw_AASM/Massimini2004_0.10-4.00Hz'

    sw = detsw(data)
    assert len(sw.events) == 15


def test_detect_slowwave_to_data():
    detsw = DetectSlowWave()
    detsw.invert = True
    sw = detsw(data)

    sw_data = sw.to_data('count')
    assert sw_data(0)[1] == 1

    sw_ptp = sw.to_data('ptp')
    assert approx(sw_ptp(0)[1]) == 63.0



"""Class to import straight text records.
"""
from logging import getLogger
from numpy import empty, float64, floor
from os import listdir
from os.path import splitext
from pathlib import Path

from .utils import DEFAULT_DATETIME

lg = getLogger(__name__)


class Text:
    """Class to read text format records. The record consists of a directory 
    containing txt files. Each file is a  channel. The first line of each file 
    is the sampling rate, and the following lines are single data points, 
    in scientific notation.
    Only supports a single sampling rate for all channels.

    Parameters
    ----------
    rec_dir : path to record directory
        the folder containing the record
        
    Notes
    -----
    Text is a very slow format for reading data. It is best to use this class
    to import the record, then to export is as a Wonambi (.won) file, and use
    that for reading.
    """
    def __init__(self, rec_dir):
        lg.info('Reading ' + str(rec_dir))
        self.filename = rec_dir
        self.hdr = self.return_hdr()
        
        # range data are absent
        self.dig_min = -0.000512 # estimated from min values
        self.dig_max = 0.000512 # estimated from max values
        self.phys_min = -800 # estimate
        self.phys_max = 800 # estimate
        
    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            the full header
        """
        foldername = Path(self.filename)
        chan_files = self.chan_files = []
        hdr = {}
        hdr['chan_name'] = []
        hdr['start_time'] = DEFAULT_DATETIME
        
        for file in listdir(self.filename):
            base, suffix = splitext(file)
            if suffix == '.txt':
                if base[-3:] == 'hyp':
                    self.hypno_file = file
                else:
                    chan_files.append(foldername / file)
                    chan_name = base[base.index('_') + 1:]
                    hdr['chan_name'].append(chan_name)
                    hdr['subj_id'] = base[:base.index('_')]
                    
        if not chan_files:
            raise FileNotFoundError('No channel found.')
            return
        
        with open(chan_files[0], 'rt') as f:
            line0 = f.readline()
            hdr['s_freq'] = int(
                    line0[line0.index('Rate:') + 5:line0.index('Hz')])
            
            for i, _ in enumerate(f):
                pass
            hdr['n_samples'] = i
        
        output = (hdr['subj_id'], hdr['start_time'], hdr['s_freq'], 
                  hdr['chan_name'], hdr['n_samples'], hdr)
        
        return output        
    
    def return_dat(self, chan, begsam, endsam):
        """Return the data as 2D numpy.ndarray.

        Parameters
        ----------
        chan : list of int
            index (indices) of the channels to read
        begsam : int
            index of the first sample (inclusively)
        endsam : int
            index of the last sample (exclusively)

        Returns
        -------
        numpy.ndarray
            A 2d matrix, with dimension chan X samples.
        """
        #n_sam = self.hdr[4]
        interval = endsam - begsam
        dat = empty((len(chan), interval))
        
        #beg_block = floor((begsam / n_sam) * n_block)
        #end_block = floor((endsam / n_sam) * n_block)
                
        for i, chan in enumerate(chan):
            k = 0
            
            with open(self.chan_files[chan], 'rt') as f:
                f.readline()
                
                for j, datum in enumerate(f):
                    
                    if begsam <= j + 1 < endsam:
                        dat[i, k] = float64(datum)
                        k += 1
                        if k == interval:
                            break
        
        # calibration
        phys_range = self.phys_max - self.phys_min
        dig_range = self.dig_max - self.dig_min
        gain = phys_range / dig_range
        dat *= gain
        
        return dat

    def return_markers(self):
        """There are no markers in this format.
        """
        return []

    
#==============================================================================
# def split_file(filepath, lines_per_file=100):
#     """splits file at `filepath` into sub-files of length `lines_per_file`
#     """
#     lpf = lines_per_file
#     path, filename = split(filepath)
#     with open(filepath, 'r') as r:
#         name, ext = splitext(filename)
#         try:
#             w = open(join(path, '{}_{}{}'.format(name, 0, ext)), 'w')
#             for i, line in enumerate(r):
#                 if not i % lpf:
#                     #possible enhancement: don't check modulo lpf on each pass
#                     #keep a counter variable, and reset on each checkpoint lpf.
#                     w.close()
#                     filename = join(path,
#                                     name,
#                                     '{}{}'.format(i, ext))
#                     w = open(filename, 'w')
#                 w.write(line)
#         finally:
#             w.close()
#==============================================================================



from datetime import datetime
from numpy import c_, NaN, empty, memmap, float64
from pathlib import Path
from scipy.io import loadmat

from .utils import (read_hdf5_str,
                    read_hdf5_chan_name,
                    DEFAULT_DATETIME,
                    )
from ..utils import MissingDependency

try:
    from h5py import File
except ImportError as err:
    File = MissingDependency(err)


class EEGLAB:
    def __init__(self, filename):
        self.filename = Path(filename).resolve()

    def return_hdr(self):
        """
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header
        """
        self.fdtfile = None

        try:
            self.EEG = loadmat(str(self.filename), struct_as_record=False,
                               squeeze_me=True)['EEG']
            self.hdf5 = False

        except NotImplementedError:
            self.hdf5 = True

        if not self.hdf5:
            self.s_freq = self.EEG.srate
            chan_name = [chan.labels for chan in self.EEG.chanlocs]
            n_samples = self.EEG.pnts

            if isinstance(self.EEG.subject, str):
                subj_id = self.EEG.subject
            else:
                subj_id = ''
            try:
                start_time = datetime(*self.EEG.etc.T0)
            except AttributeError:
                start_time = DEFAULT_DATETIME

            if isinstance(self.EEG.datfile, str):
                self.fdtfile = self.EEG.datfile
            else:
                self.data = self.EEG.data

        else:

            with File(self.filename) as f:
                EEG = f['EEG']
                self.s_freq = EEG['srate'].value.item()
                chan_name = read_hdf5_chan_name(f, EEG['chanlocs']['labels'])
                n_samples = int(EEG['pnts'].value.item())

                subj_id = read_hdf5_str(EEG['subject'])
                try:
                    start_time = datetime(*EEG['etc']['T0'])
                except ValueError:
                    start_time = DEFAULT_DATETIME

                datfile = read_hdf5_str(EEG['datfile'])
                if datfile == '':
                    self.data = EEG['data'].value.T  # for some reason, you need to transpose this
                else:
                    self.fdtfile = datfile

        if self.fdtfile is not None:
            memshape = (len(chan_name), int(n_samples))
            memmap_file = self.filename.parent / self.fdtfile
            if not memmap_file.exists():
                renamed_memmap_file = self.filename.with_suffix('.fdt')
                if not renamed_memmap_file.exists():
                    raise FileNotFoundError(f'No file {memmap_file} or {renamed_memmap_file}')
                else:
                    memmap_file = renamed_memmap_file

            self.data = memmap(str(memmap_file), 'float32', mode='c', shape=memshape, order='F')

        return subj_id, start_time, self.s_freq, chan_name, n_samples, {}

    def return_dat(self, chan, begsam, endsam):
        n_samples = self.data.shape[1]

        dat = self.data[:, max((begsam, 0)):min((endsam, n_samples))].astype(float64)

        if begsam < 0:

            pad = empty((dat.shape[0], 0 - begsam))
            pad.fill(NaN)
            dat = c_[pad, dat]

        if endsam >= n_samples:

            pad = empty((dat.shape[0], endsam - n_samples))
            pad.fill(NaN)
            dat = c_[dat, pad]

        return dat[chan, :]

    def return_markers(self):
        markers = []

        if self.hdf5:
            from h5py import File
            with File(self.filename) as f:
                for evt, latency in zip(f['EEG']['event']['type'], f['EEG']['event']['latency']):
                    mrk_t = (f[latency[0]][0, 0] - 1) / self.s_freq

                    markers.append({
                        'name': str(f[evt[0]][0, 0]),
                        'start': mrk_t,
                        'end': mrk_t,
                        })
        else:

            for event in self.EEG.event:
                markers.append({
                    'name': str(event.type),
                    'start': (event.latency - 1) / self.s_freq,
                    'end': (event.latency - 1) / self.s_freq,
                })

        return markers



from numpy.testing import assert_array_almost_equal

from wonambi import Dataset

from .paths import ktlx_file


def test_xltek_data():
    d = Dataset(ktlx_file)
    data = d.read_data(begsam=223380, endsam=223381, chan=('Fz', ))
    assert_array_almost_equal(data.data[0][0, 0], -2021.171532)


def test_xltek_marker():
    d = Dataset(ktlx_file)
    markers = d.read_markers()
    assert markers[0]['name'] == 'Gain/Filter change (-unknown-)'
    assert markers[-1]['end'] == 1052.1



from vispy.visuals import Visual
from vispy.scene.visuals import create_visual_node
from numpy import array, float32, r_
from vispy.gloo import VertexBuffer

VERT_SHADER = """
#version 120
varying vec3 v_position;
varying vec4 v_color;
varying vec3 v_normal;

void main() {
    v_position = $a_position;
    v_normal = $a_normal;
    v_color = $a_color;
    gl_Position = $transform(vec4($a_position, 1));
}
"""

FRAG_SHADER = """
#version 120
varying vec3 v_position;
varying vec4 v_color;
varying vec3 v_normal;

void main() {
    float p = dot(v_normal, normalize(vec3($u_light_position)));

    vec4 color = v_color;
    color.rgb = color.rgb * ($u_ambient + $u_diffuse * abs(p));

    gl_FragColor = color;
}
"""


class SurfaceMeshVisual(Visual):
    def __init__(self, meshdata):
        Visual.__init__(self, VERT_SHADER, FRAG_SHADER)

        v = meshdata.get_vertices(indexed='faces').astype(float32)
        self._vertices = VertexBuffer(v)

        v_norm = meshdata.get_vertex_normals(indexed='faces').astype(float32)
        self._normals = VertexBuffer(v_norm)

        v_col = meshdata.get_vertex_colors(indexed='faces').astype(float32)
        self._colors = VertexBuffer(v_col)

        self.set_light(position=(1., 0., 0.),
                       ambient=.2,
                       diffuse=.8,
                       )
        self._draw_mode = 'triangles'
        # self.set_gl_state('opaque', depth_test=True, cull_face=True)
        self.set_gl_state('translucent', depth_test=True, cull_face=False, blend=True, blend_func=('src_alpha', 'one_minus_src_alpha'))

    def _prepare_transforms(self, view):
        view.view_program.vert['transform'] = view.get_transform()

    def _prepare_draw(self, view):
        self.shared_program.vert['a_position'] = self._vertices
        self.shared_program.vert['a_normal'] = self._normals
        self.shared_program.vert['a_color'] = self._colors

    def set_light(self, position=None, ambient=None, diffuse=None):
        if position is not None:
            self.shared_program.frag['u_light_position'] = position
        if ambient is not None:
            self.shared_program.frag['u_ambient'] = ambient
        if diffuse is not None:
            self.shared_program.frag['u_diffuse'] = diffuse
        self.update()


SurfaceMesh = create_visual_node(SurfaceMeshVisual)



from logging import getLogger

from ..graphoelement import Ripple

lg = getLogger(__name__)


class DetectRipple:
    def __init__(self):
        """Design ripple detection.
        Ripples were detected as follows.
        First, data were filtered between 80–100 Hz (two-pass FIR bandpass filter, order = 3 cycles of the low frequency cut-off),
        and only artifact-free data from NREM sleep stages 2–4 were used for event detection.
        Second, the r.m.s. signal was calculated for the filtered signal using
        a moving average of 20 ms, and
        the ripple amplitude criterion was defined as the 99% percentile of
        RMS values. Third, whenever the signal exceeded this threshold
        for a minimum of 38 ms (encompassing ~3 cycles at 80 Hz) a ripple event
        was detected. In addition, we required at least three discrete peaks or
        three discrete troughs to occur in the raw signal segment corresponding
        to the above-threshold RMS segment. This was accomplished by identifying
        local maxima or minima in the respective raw signal segments after
        applying a one-pass moving average filter including the two adjacent
        data points.
        """
        pass

    def __call__(self, data):
        return Ripple()



from numpy import isnan

from wonambi import Dataset

from .paths import (eeglab_1_file,
                    eeglab_2_file,
                    eeglab_hdf5_1_file,
                    eeglab_hdf5_2_file,
                    )


def test_ioeeg_eeglab():

    d1 = Dataset(eeglab_1_file)
    d2 = Dataset(eeglab_2_file)

    data1 = d1.read_data()
    data2 = d2.read_data()

    assert data1.data[0][0, 0] == data2.data[0][0, 0]

    assert len(d1.read_markers()) == 2


def test_ioeeg_eeglab_begsam():

    d1 = Dataset(eeglab_1_file)
    data = d1.read_data(begsam=-10, endsam=1)
    assert isnan(data.data[0][0, 0])


def test_ioeeg_eeglab_hdf5():

    d1 = Dataset(eeglab_hdf5_1_file)
    d2 = Dataset(eeglab_hdf5_2_file)

    data1 = d1.read_data(begsam=100, endsam=200)
    data2 = d2.read_data(begsam=100, endsam=200)

    assert data1.data[0][0, 0] == data2.data[0][0, 0]

    assert len(d1.read_markers()) == 2



"""Class to import ABF2.
Adapted from axonrawio.py in python-neo. Strongly simplified.
"""
from datetime import datetime, timedelta
from numpy import c_, empty, float64, NaN, memmap, dtype, newaxis, array
from os import SEEK_SET
from struct import unpack, calcsize

from .utils import DEFAULT_DATETIME

BLOCKSIZE = 512


class Abf:
    """Class to read abf file.
    Only for ABF2, when the data has no gaps (no episodes).

    Parameters
    ----------
    filename : path to file
        the name of the filename with extension .won
    """
    def __init__(self, filename):
        self.filename = filename

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            the full header
        """
        with self.filename.open('br') as f:
            orig = _read_header(f)

        assert orig['protocol']['nOperationMode'] == 3, 'Only continuous no-gap recordings are supported'
        assert orig['sections']['SynchArraySection']['llNumEntries'] == 0
        assert orig['sections']['SynchArraySection']['uBlockIndex'] == 0

        # file format
        if orig['nDataFormat'] == 0:
            self.dtype = dtype('i2')
        elif orig['nDataFormat'] == 1:
            self.dtype = dtype('f4')

        chan_name = []
        offset = []
        gain = []
        for ch in orig['listADCInfo']:
            chan_name.append(ch['ADCChNames'].decode('utf-8').strip())

            # compute the gain and offset
            ch_gain = orig['protocol']['lADCResolution'] / orig['protocol']['fADCRange']
            ch_gain *= (ch['fInstrumentScaleFactor'] *
                        ch['fSignalGain'] *
                        ch['fADCProgrammableGain'])
            if ch['nTelegraphEnable'] == 1:
                ch_gain *= ch['fTelegraphAdditGain']
            gain.append(1 / ch_gain)

            offset.append(ch['fInstrumentOffset'] -
                          ch['fSignalOffset'])

        self.offset = array(offset)[:, newaxis]
        self.gain = array(gain)[:, newaxis]
        self.head = orig['sections']['DataSection']['uBlockIndex'] * BLOCKSIZE
        self.n_chan = orig['sections']['ADCSection']['llNumEntries']
        assert self.n_chan == len(chan_name)
        self.n_samples = int(orig['sections']['DataSection']['llNumEntries'] /
                             self.n_chan)

        subj_id = self.filename.stem
        try:
            start_time = (datetime.strptime(str(orig['uFileStartDate']), '%Y%m%d') +
                          timedelta(seconds=orig['uFileStartTimeMS'] / 1000))
        except ValueError:  # no time given, use placeholder
            start_time = DEFAULT_DATETIME

        s_freq = 1.e6 / orig['protocol']['fADCSequenceInterval']

        return subj_id, start_time, s_freq, chan_name, self.n_samples, orig

    def return_dat(self, chan, begsam, endsam):
        """Return the data as 2D numpy.ndarray.

        Parameters
        ----------
        chan : int or list
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, with dimension chan X samples. To save memory, the
            data are memory-mapped, and you cannot change the values on disk.

        Notes
        -----
        When asking for an interval outside the data boundaries, it returns NaN
        for those values.
        """
        data = memmap(self.filename, dtype=self.dtype, mode='r', order='F',
                      shape=(self.n_chan, self.n_samples), offset=self.head)

        dat = data[chan, max((begsam, 0)):min((endsam, self.n_samples))].astype(float64)
        dat = (dat + self.offset[chan, :]) * self.gain[chan, :]

        if begsam < 0:

            pad = empty((dat.shape[0], 0 - begsam))
            pad.fill(NaN)
            dat = c_[pad, dat]

        if endsam >= self.n_samples:

            pad = empty((dat.shape[0], endsam - self.n_samples))
            pad.fill(NaN)
            dat = c_[dat, pad]

        return dat

    def return_markers(self):
        """I don't know if the ABF contains markers at all.
        """
        return []


def _read_header(fid):
    """Based on neo/rawio/axonrawio.py, but I only kept of data with no-gaps
    and in one segment.
    """
    fid.seek(0, SEEK_SET)
    fFileSignature = fid.read(4)
    assert fFileSignature == b'ABF2', 'only format ABF2 is currently supported'

    header = {}
    for key, offset, fmt in headerDescriptionV2:
        fid.seek(0 + offset, SEEK_SET)
        val = unpack(fmt, fid.read(calcsize(fmt)))
        if len(val) == 1:
            header[key] = val[0]
        else:
            header[key] = val

    # sections
    sections = {}
    for s, sectionName in enumerate(sectionNames):
        fid.seek(76 + s * 16)
        uBlockIndex, uBytes, llNumEntries = unpack('IIl', fid.read(calcsize('IIl')))
        sections[sectionName] = {}
        sections[sectionName]['uBlockIndex'] = uBlockIndex
        sections[sectionName]['uBytes'] = uBytes
        sections[sectionName]['llNumEntries'] = llNumEntries
    header['sections'] = sections

    # strings sections
    # hack for reading channels names and units
    fid.seek(sections['StringsSection']['uBlockIndex'] * BLOCKSIZE)
    big_string = fid.read(sections['StringsSection']['uBytes'])
    goodstart = -1
    for key in [b'AXENGN', b'clampex', b'Clampex', b'CLAMPEX', b'axoscope', b'Clampfit']:
        goodstart = big_string.find(key)
        if goodstart != -1:
            break
    assert goodstart != -1, 'This file does not contain clampex, axoscope or clampfit in the header'
    big_string = big_string[goodstart:]
    strings = big_string.split(b'\x00')

    # ADC sections
    header['listADCInfo'] = []
    for i in range(sections['ADCSection']['llNumEntries']):
        # read ADCInfo
        fid.seek(sections['ADCSection']['uBlockIndex'] *
                 BLOCKSIZE + sections['ADCSection']['uBytes'] * i)
        ADCInfo = _read_info_as_dict(fid, ADCInfoDescription)
        ADCInfo['ADCChNames'] = strings[ADCInfo['lADCChannelNameIndex'] - 1]
        ADCInfo['ADCChUnits'] = strings[ADCInfo['lADCUnitsIndex'] - 1]
        header['listADCInfo'].append(ADCInfo)

    # protocol sections
    fid.seek(sections['ProtocolSection']['uBlockIndex'] * BLOCKSIZE)
    header['protocol'] = _read_info_as_dict(fid, protocolInfoDescription)
    header['sProtocolPath'] = strings[header['uProtocolPathIndex'] - 1]

    # DAC sections
    header['listDACInfo'] = []
    for i in range(sections['DACSection']['llNumEntries']):
        # read DACInfo
        fid.seek(sections['DACSection']['uBlockIndex'] *
                 BLOCKSIZE + sections['DACSection']['uBytes'] * i)
        DACInfo = _read_info_as_dict(fid, DACInfoDescription)
        DACInfo['DACChNames'] = strings[DACInfo['lDACChannelNameIndex'] - 1]
        DACInfo['DACChUnits'] = strings[
            DACInfo['lDACChannelUnitsIndex'] - 1]

        header['listDACInfo'].append(DACInfo)

    """ Not present in test file. No tests, no code.
    # tags
    listTag = []
    for i in range(sections['TagSection']['llNumEntries']):
        fid.seek(sections['TagSection']['uBlockIndex'] *
                 BLOCKSIZE + sections['TagSection']['uBytes'] * i)
        tag = _read_info_as_dict(fid, TagInfoDescription)
        listTag.append(tag)

    header['listTag'] = listTag

    # EpochPerDAC  sections
    # header['dictEpochInfoPerDAC'] is dict of dicts:
    #  - the first index is the DAC number
    #  - the second index is the epoch number
    # It has to be done like that because data may not exist
    # and may not be in sorted order
    header['dictEpochInfoPerDAC'] = {}
    for i in range(sections['EpochPerDACSection']['llNumEntries']):
        #  read DACInfo
        fid.seek(sections['EpochPerDACSection']['uBlockIndex'] *
                 BLOCKSIZE +
                 sections['EpochPerDACSection']['uBytes'] * i)
        EpochInfoPerDAC = _read_info_as_dict(fid, EpochInfoPerDACDescription)
        DACNum = EpochInfoPerDAC['nDACNum']
        EpochNum = EpochInfoPerDAC['nEpochNum']
        # Checking if the key exists, if not, the value is empty
        # so we have to create empty dict to populate
        if DACNum not in header['dictEpochInfoPerDAC']:
            header['dictEpochInfoPerDAC'][DACNum] = {}

        header['dictEpochInfoPerDAC'][DACNum][EpochNum] =\
            EpochInfoPerDAC
    """

    return header


def _read_info_as_dict(fid, values):
    """Convenience function to read info in axon data to a nicely organized
    dict.
    """
    output = {}
    for key, fmt in values:
        val = unpack(fmt, fid.read(calcsize(fmt)))
        if len(val) == 1:
            output[key] = val[0]
        else:
            output[key] = val
    return output


headerDescriptionV2 = [
    ('fFileSignature', 0, '4s'),
    ('fFileVersionNumber', 4, '4b'),
    ('uFileInfoSize', 8, 'I'),
    ('lActualEpisodes', 12, 'I'),
    ('uFileStartDate', 16, 'I'),
    ('uFileStartTimeMS', 20, 'I'),
    ('uStopwatchTime', 24, 'I'),
    ('nFileType', 28, 'H'),
    ('nDataFormat', 30, 'H'),
    ('nSimultaneousScan', 32, 'H'),
    ('nCRCEnable', 34, 'H'),
    ('uFileCRC', 36, 'I'),
    ('FileGUID', 40, 'I'),
    ('uCreatorVersion', 56, 'I'),
    ('uCreatorNameIndex', 60, 'I'),
    ('uModifierVersion', 64, 'I'),
    ('uModifierNameIndex', 68, 'I'),
    ('uProtocolPathIndex', 72, 'I'),
    ]


sectionNames = [
    'ProtocolSection',
    'ADCSection',
    'DACSection',
    'EpochSection',
    'ADCPerDACSection',
    'EpochPerDACSection',
    'UserListSection',
    'StatsRegionSection',
    'MathSection',
    'StringsSection',
    'DataSection',
    'TagSection',
    'ScopeSection',
    'DeltaSection',
    'VoiceTagSection',
    'SynchArraySection',
    'AnnotationSection',
    'StatsSection',
    ]


protocolInfoDescription = [
    ('nOperationMode', 'h'),
    ('fADCSequenceInterval', 'f'),
    ('bEnableFileCompression', 'b'),
    ('sUnused1', '3s'),
    ('uFileCompressionRatio', 'I'),
    ('fSynchTimeUnit', 'f'),
    ('fSecondsPerRun', 'f'),
    ('lNumSamplesPerEpisode', 'i'),
    ('lPreTriggerSamples', 'i'),
    ('lEpisodesPerRun', 'i'),
    ('lRunsPerTrial', 'i'),
    ('lNumberOfTrials', 'i'),
    ('nAveragingMode', 'h'),
    ('nUndoRunCount', 'h'),
    ('nFirstEpisodeInRun', 'h'),
    ('fTriggerThreshold', 'f'),
    ('nTriggerSource', 'h'),
    ('nTriggerAction', 'h'),
    ('nTriggerPolarity', 'h'),
    ('fScopeOutputInterval', 'f'),
    ('fEpisodeStartToStart', 'f'),
    ('fRunStartToStart', 'f'),
    ('lAverageCount', 'i'),
    ('fTrialStartToStart', 'f'),
    ('nAutoTriggerStrategy', 'h'),
    ('fFirstRunDelayS', 'f'),
    ('nChannelStatsStrategy', 'h'),
    ('lSamplesPerTrace', 'i'),
    ('lStartDisplayNum', 'i'),
    ('lFinishDisplayNum', 'i'),
    ('nShowPNRawData', 'h'),
    ('fStatisticsPeriod', 'f'),
    ('lStatisticsMeasurements', 'i'),
    ('nStatisticsSaveStrategy', 'h'),
    ('fADCRange', 'f'),
    ('fDACRange', 'f'),
    ('lADCResolution', 'i'),
    ('lDACResolution', 'i'),
    ('nExperimentType', 'h'),
    ('nManualInfoStrategy', 'h'),
    ('nCommentsEnable', 'h'),
    ('lFileCommentIndex', 'i'),
    ('nAutoAnalyseEnable', 'h'),
    ('nSignalType', 'h'),
    ('nDigitalEnable', 'h'),
    ('nActiveDACChannel', 'h'),
    ('nDigitalHolding', 'h'),
    ('nDigitalInterEpisode', 'h'),
    ('nDigitalDACChannel', 'h'),
    ('nDigitalTrainActiveLogic', 'h'),
    ('nStatsEnable', 'h'),
    ('nStatisticsClearStrategy', 'h'),
    ('nLevelHysteresis', 'h'),
    ('lTimeHysteresis', 'i'),
    ('nAllowExternalTags', 'h'),
    ('nAverageAlgorithm', 'h'),
    ('fAverageWeighting', 'f'),
    ('nUndoPromptStrategy', 'h'),
    ('nTrialTriggerSource', 'h'),
    ('nStatisticsDisplayStrategy', 'h'),
    ('nExternalTagType', 'h'),
    ('nScopeTriggerOut', 'h'),
    ('nLTPType', 'h'),
    ('nAlternateDACOutputState', 'h'),
    ('nAlternateDigitalOutputState', 'h'),
    ('fCellID', '3f'),
    ('nDigitizerADCs', 'h'),
    ('nDigitizerDACs', 'h'),
    ('nDigitizerTotalDigitalOuts', 'h'),
    ('nDigitizerSynchDigitalOuts', 'h'),
    ('nDigitizerType', 'h'),
    ]


ADCInfoDescription = [
    ('nADCNum', 'h'),
    ('nTelegraphEnable', 'h'),
    ('nTelegraphInstrument', 'h'),
    ('fTelegraphAdditGain', 'f'),
    ('fTelegraphFilter', 'f'),
    ('fTelegraphMembraneCap', 'f'),
    ('nTelegraphMode', 'h'),
    ('fTelegraphAccessResistance', 'f'),
    ('nADCPtoLChannelMap', 'h'),
    ('nADCSamplingSeq', 'h'),
    ('fADCProgrammableGain', 'f'),
    ('fADCDisplayAmplification', 'f'),
    ('fADCDisplayOffset', 'f'),
    ('fInstrumentScaleFactor', 'f'),
    ('fInstrumentOffset', 'f'),
    ('fSignalGain', 'f'),
    ('fSignalOffset', 'f'),
    ('fSignalLowpassFilter', 'f'),
    ('fSignalHighpassFilter', 'f'),
    ('nLowpassFilterType', 'b'),
    ('nHighpassFilterType', 'b'),
    ('fPostProcessLowpassFilter', 'f'),
    ('nPostProcessLowpassFilterType', 'c'),
    ('bEnabledDuringPN', 'b'),
    ('nStatsChannelPolarity', 'h'),
    ('lADCChannelNameIndex', 'i'),
    ('lADCUnitsIndex', 'i'),
    ]


TagInfoDescription = [
    ('lTagTime', 'i'),
    ('sComment', '56s'),
    ('nTagType', 'h'),
    ('nVoiceTagNumber_or_AnnotationIndex', 'h'),
    ]


DACInfoDescription = [
    ('nDACNum', 'h'),
    ('nTelegraphDACScaleFactorEnable', 'h'),
    ('fInstrumentHoldingLevel', 'f'),
    ('fDACScaleFactor', 'f'),
    ('fDACHoldingLevel', 'f'),
    ('fDACCalibrationFactor', 'f'),
    ('fDACCalibrationOffset', 'f'),
    ('lDACChannelNameIndex', 'i'),
    ('lDACChannelUnitsIndex', 'i'),
    ('lDACFilePtr', 'i'),
    ('lDACFileNumEpisodes', 'i'),
    ('nWaveformEnable', 'h'),
    ('nWaveformSource', 'h'),
    ('nInterEpisodeLevel', 'h'),
    ('fDACFileScale', 'f'),
    ('fDACFileOffset', 'f'),
    ('lDACFileEpisodeNum', 'i'),
    ('nDACFileADCNum', 'h'),
    ('nConditEnable', 'h'),
    ('lConditNumPulses', 'i'),
    ('fBaselineDuration', 'f'),
    ('fBaselineLevel', 'f'),
    ('fStepDuration', 'f'),
    ('fStepLevel', 'f'),
    ('fPostTrainPeriod', 'f'),
    ('fPostTrainLevel', 'f'),
    ('nMembTestEnable', 'h'),
    ('nLeakSubtractType', 'h'),
    ('nPNPolarity', 'h'),
    ('fPNHoldingLevel', 'f'),
    ('nPNNumADCChannels', 'h'),
    ('nPNPosition', 'h'),
    ('nPNNumPulses', 'h'),
    ('fPNSettlingTime', 'f'),
    ('fPNInterpulse', 'f'),
    ('nLTPUsageOfDAC', 'h'),
    ('nLTPPresynapticPulses', 'h'),
    ('lDACFilePathIndex', 'i'),
    ('fMembTestPreSettlingTimeMS', 'f'),
    ('fMembTestPostSettlingTimeMS', 'f'),
    ('nLeakSubtractADCIndex', 'h'),
    ('sUnused', '124s'),
    ]


EpochInfoPerDACDescription = [
    ('nEpochNum', 'h'),
    ('nDACNum', 'h'),
    ('nEpochType', 'h'),
    ('fEpochInitLevel', 'f'),
    ('fEpochLevelInc', 'f'),
    ('lEpochInitDuration', 'i'),
    ('lEpochDurationInc', 'i'),
    ('lEpochPulsePeriod', 'i'),
    ('lEpochPulseWidth', 'i'),
    ('sUnused', '18s'),
    ]
